# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import re

import pytest

from bumpver import v1patterns
from bumpver import v2patterns

V2_PART_PATTERN_CASES = [
    (['YYYY', 'GGGG'], "2020" , "2020"),
    (['YYYY', 'GGGG'], ""     , None),
    (['YYYY', 'GGGG'], "A020" , None),
    (['YYYY', 'GGGG'], "020"  , None),
    (['YYYY', 'GGGG'], "12020", None),
    (['YY'  , 'GG'  ], "20"   , "20"),
    (['YY'  , 'GG'  ], "3"    , "3"),
    (['YY'  , 'GG'  ], "03"   , None),
    (['YY'  , 'GG'  ], "2X"   , None),
    (['YY'  , 'GG'  ], ""     , None),
    (['0Y'  , '0G'  ], "20"   , "20"),
    (['0Y'  , '0G'  ], "03"   , "03"),
    (['0Y'  , '0G'  ], "3"    , None),
    (['0Y'  , '0G'  ], "2X"   , None),
    (['0Y'  , '0G'  ], ""     , None),
    # quarter
    (['Q'], "0", None),
    (['Q'], "1", "1"),
    (['Q'], "2", "2"),
    (['Q'], "3", "3"),
    (['Q'], "4", "4"),
    (['Q'], "5", None),
    (['Q'], "X", None),
    # months
    (['MM'], "0" , None),
    (['MM'], "01", None),
    (['MM'], "1" , "1"),
    (['MM'], "10", "10"),
    (['MM'], "12", "12"),
    (['MM'], "13", None),
    (['0M'], "00", None),
    (['0M'], "1" , None),
    (['0M'], "01", "01"),
    (['MM'], "10", "10"),
    (['MM'], "12", "12"),
    (['MM'], "13", None),
    # day of month
    (['DD'], "0" , None),
    (['DD'], "01", None),
    (['DD'], "1" , "1"),
    (['DD'], "10", "10"),
    (['DD'], "31", "31"),
    (['DD'], "32", None),
    (['0D'], "00", None),
    (['0D'], "1" , None),
    (['0D'], "01", "01"),
    (['0D'], "10", "10"),
    (['0D'], "31", "31"),
    (['0D'], "32", None),
    (['DD'], "0" , None),
    (['DD'], "01", None),
    (['DD'], "1" , "1"),
    (['DD'], "10", "10"),
    (['DD'], "31", "31"),
    (['DD'], "32", None),
    (['0D'], "00", None),
    (['0D'], "1" , None),
    (['0D'], "01", "01"),
    (['0D'], "10", "10"),
    (['0D'], "31", "31"),
    (['0D'], "32", None),
    # day of year
    (['JJJ'], "0"  , None),
    (['JJJ'], "01" , None),
    (['JJJ'], "1"  , "1"),
    (['JJJ'], "10" , "10"),
    (['JJJ'], "31" , "31"),
    (['JJJ'], "32" , "32"),
    (['JJJ'], "100", "100"),
    (['JJJ'], "365", "365"),
    (['JJJ'], "366", "366"),
    (['JJJ'], "367", None),
    (['00J'], "000", None),
    (['00J'], "01" , None),
    (['00J'], "1"  , None),
    (['00J'], "001", "001"),
    (['00J'], "010", "010"),
    (['00J'], "031", "031"),
    (['00J'], "032", "032"),
    (['00J'], "100", "100"),
    (['00J'], "365", "365"),
    (['00J'], "366", "366"),
    (['00J'], "367", None),
    # week numbers
    (['WW', 'UU'], "00", None),
    (['WW', 'UU'], "01", None),
    (['WW', 'UU'], "0" , "0"),
    (['WW', 'UU'], "1" , "1"),
    (['WW', 'UU'], "10", "10"),
    (['WW', 'UU'], "52", "52"),
    (['WW', 'UU'], "53", None),
    (['0W', '0U'], "00", "00"),
    (['0W', '0U'], "01", "01"),
    (['0W', '0U'], "0" , None),
    (['0W', '0U'], "1" , None),
    (['0W', '0U'], "10", "10"),
    (['0W', '0U'], "52", "52"),
    (['0W', '0U'], "53", None),
    (['VV'], "00", None),
    (['VV'], "01", None),
    (['VV'], "0" , None),
    (['VV'], "1" , "1"),
    (['VV'], "10", "10"),
    (['VV'], "52", "52"),
    (['VV'], "53", "53"),
    (['VV'], "54", None),
    (['0V'], "00", None),
    (['0V'], "01", "01"),
    (['0V'], "0" , None),
    (['0V'], "1" , None),
    (['0V'], "10", "10"),
    (['0V'], "52", "52"),
    (['0V'], "53", "53"),
    (['0V'], "54", None),
    (['MAJOR', 'MINOR', 'PATCH'], "0", "0"),
    (['TAG'  ], "alpha" , "alpha"),
    (['TAG'  ], "alfa"  , None),
    (['TAG'  ], "beta"  , "beta"),
    (['TAG'  ], "rc"    , "rc"),
    (['TAG'  ], "post"  , "post"),
    (['TAG'  ], "final" , "final"),
    (['TAG'  ], "latest", None),
    (['PYTAG'], "a"     , "a"),
    (['PYTAG'], "b"     , "b"),
    (['PYTAG'], "rc"    , "rc"),
    (['PYTAG'], "post"  , "post"),
    (['PYTAG'], "post"  , "post"),
    (['PYTAG'], "x"     , None),
    (['NUM'  ], "a"     , None),
    (['NUM'  ], "0"     , "0"),
    (['NUM'  ], "1"     , "1"),
    (['NUM'  ], "10"    , "10"),
]


def _compile_part_re(pattern_str):
    grouped_pattern_str = r"^(?:" + pattern_str + r")$"
    # print("\n", grouped_pattern_str)
    return re.compile(grouped_pattern_str, flags=re.MULTILINE)


@pytest.mark.parametrize("parts, testcase, expected", V2_PART_PATTERN_CASES)
def test_v2_part_patterns(parts, testcase, expected):
    for part in parts:
        part_re = _compile_part_re(v2patterns.PART_PATTERNS[part])
        match   = part_re.match(testcase)
        if match is None:
            assert expected is None
        else:
            assert match.group(0) == expected


@pytest.mark.parametrize("part_name", v2patterns.PART_PATTERNS.keys())
def test_v1_part_compilation(part_name):
    assert _compile_part_re(v2patterns.PART_PATTERNS[part_name])


@pytest.mark.parametrize("part_name", v1patterns.PART_PATTERNS.keys())
def test_v2_part_compilation(part_name):
    assert _compile_part_re(v1patterns.PART_PATTERNS[part_name])


PATTERN_PART_CASES = [
    ("pep440_pycalver", "201712.31"          , "201712.31"),
    ("pep440_pycalver", "v201712.0032"       , None),
    ("pep440_pycalver", "201712.0033-alpha"  , None),
    ("pycalver"       , "v201712.0034"       , "v201712.0034"),
    ("pycalver"       , "v201712.0035-alpha" , "v201712.0035-alpha"),
    ("pycalver"       , "v201712.0036-alpha0", None),
    ("pycalver"       , "v201712.0037-pre"   , None),  # pre not available for v1 patterns
    ("pycalver"       , "201712.38a0"        , None),
    ("pycalver"       , "201712.0039"        , None),
    ("semver"         , "1.23.456"           , "1.23.456"),
    ("calver"         , "v201712"            , "v201712"),
    ("calver"         , "v201799"            , None),  # invalid date
    ("calver"         , "201712"             , None),
    ("calver"         , "v20171"             , None),
    ("build"          , ".0012"              , ".0012"),
    ("build"          , ".11012"             , ".11012"),
    ("build"          , ".012"               , None),
    ("build"          , "11012"              , None),
    ("release"        , "-alpha"             , "-alpha"),
    ("release"        , "-beta"              , "-beta"),
    ("release"        , "-dev"               , "-dev"),
    ("release"        , "-rc"                , "-rc"),
    ("release"        , "-post"              , "-post"),
    ("release"        , "-pre"               , None),  # pre not available for v1 patterns
    ("release"        , "alpha"              , None),  # missing dash "-" prefix
]


@pytest.mark.parametrize("part_name, line, expected", PATTERN_PART_CASES)
def test_v1_re_pattern_parts(part_name, line, expected):
    part_re = _compile_part_re(v1patterns.PART_PATTERNS[part_name])
    result  = part_re.match(line)
    if result is None:
        assert expected is None, (part_name, line, result)
    else:
        result_val = result.group(0)
        assert result_val == expected, (part_name, line)


PATTERN_V1_CASES = [
    (r"v{year}.{month}.{MINOR}"      , "v2017.11.1" , "v2017.11.1"),
    (r"v{year}.{month}.{MINOR}"      , "v2017.07.12", "v2017.07.12"),
    (r"v{year}.{month_short}.{PATCH}", "v2017.11.1" , "v2017.11.1"),
    (r"v{year}.{month_short}.{PATCH}", "v2017.7.12" , "v2017.7.12"),
]


@pytest.mark.parametrize("pattern_str, line, expected", PATTERN_V1_CASES)
def test_v1_patterns(pattern_str, line, expected):
    pattern = v1patterns.compile_pattern(pattern_str)
    result  = pattern.regexp.search(line)
    if result is None:
        assert expected is None, (pattern_str, line)
    else:
        result_val = result.group(0)
        assert result_val == expected, (pattern_str, line)


PATTERN_V2_CASES = [
    ("vYYYY.0M.MINOR" , "v2017.11.1" , "v2017.11.1"),
    ("vYYYY.0M.MINOR" , "v2017.07.12", "v2017.07.12"),
    ("YYYY.MM[.PATCH]", "2017.11.1"  , "2017.11.1"),
    ("YYYY.MM[.PATCH]", "2017.7.12"  , "2017.7.12"),
    ("YYYY.MM[.PATCH]", "2017.7"     , "2017.7"),
    ("YYYY0M.BUILD"   , "201707.1000", "201707.1000"),
]


@pytest.mark.parametrize("pattern_str, line, expected", PATTERN_V2_CASES)
def test_v2_patterns(pattern_str, line, expected):
    pattern    = v2patterns.compile_pattern(pattern_str)
    result     = pattern.regexp.search(line)
    result_val = None if result is None else result.group(0)
    assert result_val == expected, (pattern_str, line, pattern.regexp.pattern)


def test_epoch_pattern():
    raw_pattern = v2patterns.normalize_pattern("v1!MAJOR.MINOR.PATCH", "{version}")
    pattern     = v2patterns.compile_pattern(raw_pattern)

    result     = pattern.regexp.search('current_version = "v1!2.1.5"')
    result_val = None if result is None else result.group(0)
    assert result_val == "v1!2.1.5"


def test_epoch_pep440_pattern():
    raw_pep440_pattern = v2patterns.normalize_pattern("v1!MAJOR.MINOR.PATCH", "{pep440_version}")
    pep440_pattern     = v2patterns.compile_pattern(raw_pep440_pattern)

    result     = pep440_pattern.regexp.search('version = "1!2.1.5"')
    result_val = None if result is None else result.group(0)
    assert result_val == "1!2.1.5"


CLI_MAIN_FIXTURE = """
@click.group()
@click.version_option(version="v201812.0123-beta")
@click.help_option()
"""


def test_pattern_escapes():
    pattern_str = 'click.version_option(version="{pycalver}")'
    pattern     = v1patterns.compile_pattern(pattern_str)
    match       = pattern.regexp.search(CLI_MAIN_FIXTURE)
    expected    = 'click.version_option(version="v201812.0123-beta")'
    assert match.group(0) == expected


CURLY_BRACE_FIXTURE = """
package_metadata = {"name": "mypackage", "version": "v201812.0123-beta"}
"""


def test_curly_escapes():
    pattern_str = 'package_metadata = {"name": "mypackage", "version": "{pycalver}"}'
    pattern     = v1patterns.compile_pattern(pattern_str)
    match       = pattern.regexp.search(CURLY_BRACE_FIXTURE)
    expected    = 'package_metadata = {"name": "mypackage", "version": "v201812.0123-beta"}'
    assert match.group(0) == expected


def test_part_field_mapping_v2():
    a_names = set(v2patterns.PATTERN_PART_FIELDS.keys())
    b_names = set(v2patterns.PART_PATTERNS.keys())

    a_extra_names = a_names - b_names
    assert not any(a_extra_names), sorted(a_extra_names)
    b_extra_names = b_names - a_names
    assert not any(b_extra_names), sorted(b_extra_names)



# This file is part of the bumpver project
# https://github.com/mbarkhau/bumpver
#
# Copyright (c) 2018-2023 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
"""Compose Regular Expressions from Patterns.

>>> pattern = compile_pattern("vYYYY0M.BUILD[-TAG]")
>>> version_info = pattern.regexp.match("v201712.0123-alpha")
>>> assert version_info.groupdict() == {
...     "year_y" : "2017",
...     "month"  : "12",
...     "bid"    : "0123",
...     "tag"    : "alpha",
... }
>>>
>>> version_info = pattern.regexp.match("201712.1234")
>>> assert version_info is None

>>> version_info = pattern.regexp.match("v201713.1234")
>>> assert version_info is None

>>> version_info = pattern.regexp.match("v201712.1234")
>>> assert version_info.groupdict() == {
...     "year_y" : "2017",
...     "month"  : "12",
...     "bid"    : "1234",
...     "tag"    : None,
... }
"""

import re
import typing as typ
import collections

from . import utils
from .patterns import RE_PATTERN_ESCAPES
from .patterns import Pattern

# NOTE (mb 2020-09-17): For patterns with different options '(AAA|BB|C)', the
#   patterns with more digits should be first/left of those with fewer digits:
#
#       good: (?:1[0-2]|[1-9])
#       bad:  (?:[1-9]|1[0-2])
#
#   This ensures that the longest match is done for a pattern.
#
#   This implies that patterns for smaller numbers sometimes must be right of
#   those for larger numbers. To be consistent we use this ordering not
#   sometimes but always (even though in theory it wouldn't matter):
#
#       good: (?:3[0-1]|[1-2][0-9]|[1-9])
#       bad:  (?:[1-2][0-9]|3[0-1]|[1-9])


PART_PATTERNS = collections.OrderedDict(
    [
        # Based on calver.org
        ('YYYY', r"[1-9][0-9]{3}"),
        ('YY'  , r"[1-9][0-9]?"),
        ('0Y'  , r"[0-9]{2}"),
        ('GGGG', r"[1-9][0-9]{3}"),
        ('GG'  , r"[1-9][0-9]?"),
        ('0G'  , r"[0-9]{2}"),
        ('Q'   , r"[1-4]"),
        ('MM'  , r"1[0-2]|[1-9]"),
        ('0M'  , r"1[0-2]|0[1-9]"),
        ('DD'  , r"3[0-1]|[1-2][0-9]|[1-9]"),
        ('0D'  , r"3[0-1]|[1-2][0-9]|0[1-9]"),
        ('JJJ' , r"36[0-6]|3[0-5][0-9]|[1-2][0-9][0-9]|[1-9][0-9]|[1-9]"),
        ('00J' , r"36[0-6]|3[0-5][0-9]|[1-2][0-9][0-9]|0[1-9][0-9]|00[1-9]"),
        # week numbering parts
        ('WW', r"5[0-2]|[1-4][0-9]|[0-9]"),
        ('0W', r"5[0-2]|[0-4][0-9]"),
        ('UU', r"5[0-2]|[1-4][0-9]|[0-9]"),
        ('0U', r"5[0-2]|[0-4][0-9]"),
        ('VV', r"5[0-3]|[1-4][0-9]|[1-9]"),
        ('0V', r"5[0-3]|[1-4][0-9]|0[1-9]"),
        # non calver parts
        ('MAJOR'  , r"[0-9]+"),
        ('MINOR'  , r"[0-9]+"),
        ('PATCH'  , r"[0-9]+"),
        ('BUILD'  , r"[0-9]+"),
        ('BLD'    , r"[1-9][0-9]*"),
        ('TAG'    , r"preview|final|dev|alpha|beta|post|rc"),
        ('PYTAG'  , r"dev|post|rc|a|b"),
        ('GITHASH', r"\.[0-9]+\+.*"),
        ('HEXHASH', r"[0-9a-f]+"),
        ('NUM'    , r"[0-9]+"),
        ('INC0'   , r"[0-9]+"),
        ('INC1'   , r"[1-9][0-9]*"),
    ]
)


PATTERN_PART_FIELDS = {
    'YYYY'   : 'year_y',
    'YY'     : 'year_y',
    '0Y'     : 'year_y',
    'GGGG'   : 'year_g',
    'GG'     : 'year_g',
    '0G'     : 'year_g',
    'Q'      : 'quarter',
    'MM'     : 'month',
    '0M'     : 'month',
    'DD'     : 'dom',
    '0D'     : 'dom',
    'JJJ'    : 'doy',
    '00J'    : 'doy',
    'MAJOR'  : 'major',
    'MINOR'  : 'minor',
    'PATCH'  : 'patch',
    'BUILD'  : 'bid',
    'BLD'    : 'bid',
    'TAG'    : 'tag',
    'PYTAG'  : 'pytag',
    'GITHASH': 'githash',
    'HEXHASH': 'hexhash',
    'NUM'    : 'num',
    'INC0'   : 'inc0',
    'INC1'   : 'inc1',
    'WW'     : 'week_w',
    '0W'     : 'week_w',
    'UU'     : 'week_u',
    '0U'     : 'week_u',
    'VV'     : 'week_v',
    '0V'     : 'week_v',
}


PEP440_PART_SUBSTITUTIONS = {
    '0W'   : "WW",
    '0U'   : "UU",
    '0V'   : "VV",
    '0M'   : "MM",
    '0D'   : "DD",
    '00J'  : "JJJ",
    'BUILD': "BLD",
    'TAG'  : "PYTAG",
}


FieldValue = typ.Union[str, int]


def _fmt_num(val: FieldValue) -> str:
    return str(val)


def _fmt_bld(val: FieldValue) -> str:
    return str(int(val))


def _fmt_yy(year_y: FieldValue) -> str:
    return str(int(str(year_y)[-2:]))


def _fmt_0y(year_y: FieldValue) -> str:
    return f"{int(str(year_y)[-2:]):02}"


def _fmt_gg(year_g: FieldValue) -> str:
    return str(int(str(year_g)[-2:]))


def _fmt_0g(year_g: FieldValue) -> str:
    return f"{int(str(year_g)[-2:]):02}"


def _fmt_0m(month: FieldValue) -> str:
    return f"{int(month):02}"


def _fmt_0d(dom: FieldValue) -> str:
    return f"{int(dom):02}"


def _fmt_00j(doy: FieldValue) -> str:
    return f"{int(doy):03}"


def _fmt_0w(week_w: FieldValue) -> str:
    return f"{int(week_w):02}"


def _fmt_0u(week_u: FieldValue) -> str:
    return f"{int(week_u):02}"


def _fmt_0v(week_v: FieldValue) -> str:
    return f"{int(week_v):02}"


FormatterFunc = typ.Callable[[FieldValue], str]


PART_FORMATS: typ.Dict[str, FormatterFunc] = {
    'YYYY'   : _fmt_num,
    'YY'     : _fmt_yy,
    '0Y'     : _fmt_0y,
    'GGGG'   : _fmt_num,
    'GG'     : _fmt_gg,
    '0G'     : _fmt_0g,
    'Q'      : _fmt_num,
    'MM'     : _fmt_num,
    '0M'     : _fmt_0m,
    'DD'     : _fmt_num,
    '0D'     : _fmt_0d,
    'JJJ'    : _fmt_num,
    '00J'    : _fmt_00j,
    'MAJOR'  : _fmt_num,
    'MINOR'  : _fmt_num,
    'PATCH'  : _fmt_num,
    'BUILD'  : _fmt_num,
    'BLD'    : _fmt_bld,
    'TAG'    : _fmt_num,
    'PYTAG'  : _fmt_num,
    'GITHASH': _fmt_num,
    'HEXHASH': _fmt_num,
    'NUM'    : _fmt_num,
    'INC0'   : _fmt_num,
    'INC1'   : _fmt_num,
    'WW'     : _fmt_num,
    '0W'     : _fmt_0w,
    'UU'     : _fmt_num,
    '0U'     : _fmt_0u,
    'VV'     : _fmt_num,
    '0V'     : _fmt_0v,
}


def _convert_to_pep440(version_pattern: str) -> str:
    # NOTE (mb 2020-09-20): This does not support some
    #   corner cases as specified in PEP440, in particular
    #   related to post and dev releases.

    pep440_pattern = version_pattern

    if pep440_pattern.startswith("v"):
        pep440_pattern = pep440_pattern[1:]

    pep440_pattern = pep440_pattern.replace(r"\[", "")
    pep440_pattern = pep440_pattern.replace(r"\]", "")

    pep440_pattern, _ = re.subn(r"[^a-zA-Z0-9\.\!\[\]]", "", pep440_pattern)

    part_names = list(PATTERN_PART_FIELDS.keys())
    part_names.sort(key=len, reverse=True)

    for part_name in part_names:
        if part_name not in version_pattern:
            continue
        if part_name not in PEP440_PART_SUBSTITUTIONS:
            continue

        substitution = PEP440_PART_SUBSTITUTIONS[part_name]
        if substitution in pep440_pattern:
            continue

        is_numerical_part = part_name not in ('TAG', 'PYTAG')
        if is_numerical_part:
            part_index              = pep440_pattern.find(part_name)
            is_zero_truncation_part = part_index == 0 or pep440_pattern[part_index - 1] == "."
            if is_zero_truncation_part:
                pep440_pattern = pep440_pattern.replace(part_name, substitution)
        else:
            pep440_pattern = pep440_pattern.replace(part_name, substitution)

    # PYTAG and NUM must be adjacent and also be the last (optional) part
    if 'PYTAGNUM' not in pep440_pattern:
        pep440_pattern = pep440_pattern.replace("PYTAG", "")
        pep440_pattern = pep440_pattern.replace("NUM"  , "")
        pep440_pattern = pep440_pattern.replace("[]"   , "")
        pep440_pattern += "[PYTAGNUM]"

    return pep440_pattern


def normalize_pattern(version_pattern: str, raw_pattern: str) -> str:
    normalized_pattern = raw_pattern
    if "{version}" in raw_pattern:
        normalized_pattern = normalized_pattern.replace("{version}", version_pattern)

    if "{pep440_version}" in normalized_pattern:
        pep440_version_pattern = _convert_to_pep440(version_pattern)
        normalized_pattern     = normalized_pattern.replace("{pep440_version}", pep440_version_pattern)

    return normalized_pattern


SortKey         = typ.Tuple[int, int]
PostitionedPart = typ.Tuple[int, int, str]


def _iter_part_patterns(pattern: str) -> typ.Iterator[typ.Tuple[SortKey, PostitionedPart]]:
    used_fields: typ.Set[str] = set()
    for part_name, part_pattern in PART_PATTERNS.items():
        end_idx = 0
        while True:
            start_idx = pattern.find(part_name, end_idx)
            if start_idx < 0:
                break

            field = PATTERN_PART_FIELDS[part_name]
            if field in used_fields:
                named_part_pattern = f"(?P<{field}_{len(used_fields)}>{part_pattern})"
            else:
                named_part_pattern = f"(?P<{field}>{part_pattern})"
            used_fields.add(field)

            end_idx         = start_idx + len(part_name)
            sort_key        = (-end_idx, -len(part_name))
            positioned_part = (start_idx, end_idx, named_part_pattern)
            yield (sort_key, positioned_part)


def _replace_pattern_parts(pattern: str) -> str:
    # The pattern is escaped, so that everything besides the format
    # string variables is treated literally.
    while True:
        new_pattern, _n = re.subn(r"([^\\]|^)\[", r"\1(?:", pattern)
        new_pattern, _m = re.subn(r"([^\\]|^)\]", r"\1)?" , new_pattern)
        pattern = new_pattern
        if _n + _m == 0:
            break

    part_patterns_by_index: typ.Dict[SortKey, PostitionedPart] = dict(_iter_part_patterns(pattern))

    # NOTE (mb 2020-09-17): The sorting is done so that we process items:
    #   - right before left
    #   - longer before shorter
    last_start_idx = len(pattern) + 1
    result_pattern = pattern
    for _, (start_idx, end_idx, named_part_pattern) in sorted(part_patterns_by_index.items()):
        if end_idx <= last_start_idx:
            result_pattern = (
                result_pattern[:start_idx] + named_part_pattern + result_pattern[end_idx:]
            )
            last_start_idx = start_idx

    return result_pattern


def _compile_pattern_re(normalized_pattern: str) -> typ.Pattern[str]:
    escaped_pattern = normalized_pattern
    for char, escaped in RE_PATTERN_ESCAPES:
        # [] braces are used for optional parts, such as [-TAG]/[-beta]
        # and need to be escaped manually.
        is_semantic_char = char in "[]\\"
        if not is_semantic_char:
            # escape it so it is a literal in the re pattern
            escaped_pattern = escaped_pattern.replace(char, escaped)

    pattern_str = _replace_pattern_parts(escaped_pattern)
    return re.compile(pattern_str)


@utils.memo
def compile_pattern(version_pattern: str, raw_pattern: typ.Optional[str] = None) -> Pattern:
    _raw_pattern       = version_pattern if raw_pattern is None else raw_pattern
    normalized_pattern = normalize_pattern(version_pattern, _raw_pattern)
    regexp             = _compile_pattern_re(normalized_pattern)
    return Pattern(version_pattern, normalized_pattern, regexp)


def compile_patterns(version_pattern: str, raw_patterns: typ.List[str]) -> typ.List[Pattern]:
    return [compile_pattern(version_pattern, raw_pattern) for raw_pattern in raw_patterns]



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import io
import os
import re
import time
import shlex
import shutil
import logging
import datetime as dt
import subprocess as sp
from test import util

import pytest
from click.testing import CliRunner

from bumpver import cli
from bumpver import config
from bumpver import pathlib as pl
from bumpver import v2patterns

# pylint:disable=redefined-outer-name ; pytest fixtures
# pylint:disable=protected-access ; allowed for test code
# pylint:disable=unused-argument ; allowed for test code
# pylint:disable=too-many-lines ; allowed for test code


README_TEXT_FIXTURE = """
        Hello World v2017.1002-alpha !
        [aka. 2017.1002a0 !]
        Hello World v201707.1002-alpha !
        [aka. 201707.1002a0 !]
"""


SETUP_CFG_FIXTURE = """
[metadata]
license_file = LICENSE

[bdist_wheel]
universal = 1
"""

CALVER_TOML_FIXTURE = """
"""

PYPROJECT_TOML_FIXTURE = """
[build-system]
requires = ["setuptools", "wheel"]
"""

ENV = {
    'GIT_AUTHOR_NAME'    : "bumpver_tester",
    'GIT_COMMITTER_NAME' : "bumpver_tester",
    'GIT_AUTHOR_EMAIL'   : "bumpver_tester@nowhere.com",
    'GIT_COMMITTER_EMAIL': "bumpver_tester@nowhere.com",
    'HGUSER'             : "bumpver_tester",
    'PATH'               : os.environ['PATH'],
}


def shell(*cmd):
    return sp.check_output(cmd, env=ENV)


ECHO_CAPLOG = os.getenv('ECHO_CAPLOG') == "1"


def _debug_records(caplog):
    if ECHO_CAPLOG:
        print()
        for record in caplog.records:
            print(record)


@pytest.fixture
def runner(tmpdir):
    runner   = CliRunner(env=ENV)
    orig_cwd = os.getcwd()

    _debug = 0
    if _debug:
        tmpdir = pl.Path("..") / "tmp_test_bumpver_project"
        if tmpdir.exists():
            time.sleep(0.2)
            shutil.rmtree(str(tmpdir))
        tmpdir.mkdir()

    os.chdir(str(tmpdir))

    yield runner

    os.chdir(orig_cwd)

    if not _debug:
        shutil.rmtree(str(tmpdir))


def test_help(runner):
    result = runner.invoke(cli.cli, ['--help', "-vv"])
    assert result.exit_code == 0
    assert "update " in result.output
    assert "test " in result.output
    assert "init " in result.output
    assert "show " in result.output


def test_version(runner):
    result = runner.invoke(cli.cli, ['--version', "-vv"])
    assert result.exit_code == 0
    assert " version 20" in result.output
    pattern = v2patterns.compile_pattern("YYYY.BUILD[-TAG]")
    match   = pattern.regexp.search(result.output)
    assert match


def test_show_env(runner):
    _add_project_files("README.md", "setup.cfg")

    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    result = runner.invoke(cli.cli, ['show', "-e"])
    assert result.exit_code == 0
    assert "TAG=alpha" in result.output


def test_incr_default(runner):
    old_version = "v201709.1004-alpha"

    cmd    = ['test', "-vv", "--pin-date", "--tag", "beta", old_version, "{pycalver}"]
    result = runner.invoke(cli.cli, cmd)
    assert result.exit_code == 0
    assert "Version: v201709.1005-beta\n" in result.output

    old_version = "v2017.1004-alpha"

    cmd    = ['test', "-vv", "--pin-date", "--tag", "beta", old_version, "v{year}{build}{release}"]
    result = runner.invoke(cli.cli, cmd)
    assert result.exit_code == 0
    assert "Version: v2017.1005-beta\n" in result.output

    cmd    = ['test', "-vv", "--pin-date", "--tag", "beta", old_version, "vYYYY.BUILD[-TAG]"]
    result = runner.invoke(cli.cli, cmd)
    assert result.exit_code == 0
    assert "Version: v2017.1005-beta\n" in result.output


def test_incr_pin_date(runner):
    old_version = "v2017.1999-alpha"
    pattern     = "vYYYY.BUILD[-TAG]"
    result      = runner.invoke(cli.cli, ['test', "-vv", "--pin-date", old_version, pattern])
    assert result.exit_code == 0
    assert "Version: v2017.22000-alpha\n" in result.output


@pytest.mark.parametrize(
    "version_pattern, old_version, new_version",
    [
        ("vYYYY.INC0[-PATCH]", "v2017.0", "v2017.0-1"),
        ("vYYYY.INC0[-PATCH]", "v2017.0-1", "v2017.0-2"),
        ("vYYYY.INC1[-PATCH]", "v2017.1", "v2017.1-1"),
        ("vYYYY.INC1[-PATCH]", "v2017.1-1", "v2017.1-2"),
    ],
)
def test_incr_pin_increments(runner, version_pattern, old_version, new_version):
    result = runner.invoke(
        cli.cli,
        [
            'test',
            "-vv",
            "--pin-increments",
            "--patch",
            "--date",
            "2017-12-01",
            old_version,
            version_pattern,
        ],
    )
    assert result.exit_code == 0
    assert f"Version: {new_version}\n" in result.output


def test_incr_semver(runner):
    semver_patterns = [
        "{semver}",
        "{MAJOR}.{MINOR}.{PATCH}",
        "MAJOR.MINOR.PATCH",
    ]

    for semver_pattern in semver_patterns:
        old_version = "0.1.0"
        new_version = "0.1.1"

        result = runner.invoke(cli.cli, ['test', "-vv", "--patch", old_version, semver_pattern])
        assert result.exit_code == 0
        assert f"Version: {new_version}\n" in result.output

        old_version = "0.1.1"
        new_version = "0.2.0"

        result = runner.invoke(cli.cli, ['test', "-vv", "--minor", old_version, semver_pattern])
        assert result.exit_code == 0
        assert f"Version: {new_version}\n" in result.output

        old_version = "0.1.1"
        new_version = "1.0.0"

        result = runner.invoke(cli.cli, ['test', "-vv", "--major", old_version, semver_pattern])
        assert result.exit_code == 0
        assert f"Version: {new_version}\n" in result.output


def test_incr_semver_invalid(runner, caplog):
    pattern = "vYYYY.BUILD[-TAG]"
    result  = runner.invoke(cli.cli, ['test', "-vv", "0.1.1", pattern, "--patch"])
    assert result.exit_code == 1
    assert len(caplog.records) > 0
    log_msg = caplog.records[0].message
    assert "--patch is not applicable to pattern" in log_msg
    assert "to pattern 'vYYYY.BUILD[-TAG]'" in log_msg


def test_incr_to_beta(runner):
    pattern     = "vYYYY.BUILD[-TAG]"
    old_version = "v2017.1999-alpha"
    new_version = dt.datetime.utcnow().strftime("v%Y.22000-beta")

    result = runner.invoke(cli.cli, ['test', "-vv", old_version, pattern, "--tag", "beta"])
    assert result.exit_code == 0
    assert f"Version: {new_version}\n" in result.output


def test_incr_to_final(runner, caplog):
    pattern     = "vYYYY.BUILD[-TAG]"
    old_version = "v2017.1999-alpha"
    new_version = dt.datetime.utcnow().strftime("v%Y.22000")

    result = runner.invoke(cli.cli, ['test', "-vv", old_version, pattern, "--tag", "final"])
    _debug_records(caplog)
    assert result.exit_code == 0
    assert f"Version: {new_version}\n" in result.output


SEMVER = "MAJOR.MINOR.PATCH[PYTAGNUM]"


def test_incr_tag(runner):
    old_version = "0.1.0"
    new_version = "0.1.1b0"

    result = runner.invoke(
        cli.cli, ['test', "-vv", old_version, SEMVER, "--patch", "--tag", "beta"]
    )
    assert result.exit_code == 0
    assert f"Version: {new_version}\n" in result.output


def test_dev_tag(runner):
    old_version = "0.1.0"
    new_version = "0.1.1dev0"

    result = runner.invoke(cli.cli, ['test', "-vv", old_version, SEMVER, "--patch", "--tag", "dev"])
    assert result.exit_code == 0
    assert f"Version: {new_version}\n" in result.output


def test_incr_tag_num(runner):
    old_version = "0.1.0b0"
    new_version = "0.1.0b1"

    result = runner.invoke(cli.cli, ['test', "-vv", old_version, SEMVER, "--tag-num"])
    assert result.exit_code == 0
    assert f"Version: {new_version}\n" in result.output


def test_incr_invalid(runner):
    pattern     = "vYYYY.BUILD[-TAG]"
    old_version = "v2017.1999-alpha"

    result = runner.invoke(cli.cli, ['test', "-vv", old_version, pattern, "--tag", "alfa"])
    assert result.exit_code == 1


def _add_project_files(*files):
    if "README.md" in files:
        with pl.Path("README.md").open(mode="wt", encoding="utf-8") as fobj:
            fobj.write(README_TEXT_FIXTURE)

    if "setup.cfg" in files:
        with pl.Path("setup.cfg").open(mode="wt", encoding="utf-8") as fobj:
            fobj.write(SETUP_CFG_FIXTURE)

    if "pycalver.toml" in files:
        with pl.Path("pycalver.toml").open(mode="wt", encoding="utf-8") as fobj:
            fobj.write(CALVER_TOML_FIXTURE)

    if "pyproject.toml" in files:
        with pl.Path("pyproject.toml").open(mode="wt", encoding="utf-8") as fobj:
            fobj.write(PYPROJECT_TOML_FIXTURE)

    if "bumpver.toml" in files:
        with pl.Path("bumpver.toml").open(mode="wt", encoding="utf-8") as fobj:
            fobj.write(CALVER_TOML_FIXTURE)


def _update_config_val(filename, **kwargs):
    with io.open(filename, mode="r", encoding="utf-8") as fobj:
        old_cfg_text = fobj.read()

    new_cfg_text = old_cfg_text
    for key, val in kwargs.items():
        replacement = f"{key} = {val}"
        if replacement not in new_cfg_text:
            pattern      = fr"^{key} = .*$"
            new_cfg_text = re.sub(pattern, replacement, new_cfg_text, flags=re.MULTILINE)
            assert old_cfg_text != new_cfg_text

    with io.open(filename, mode="w", encoding="utf-8") as fobj:
        fobj.write(new_cfg_text)


def test_nocfg(runner, caplog):
    _add_project_files("README.md")
    result = runner.invoke(cli.cli, ['show', "-vv"])
    assert result.exit_code == 1
    expected_msg = "Could not parse configuration. Perhaps try 'bumpver init'."
    _debug_records(caplog)
    assert any(expected_msg in r.message for r in caplog.records)


def test_novcs_nocfg_init(runner, caplog):
    _add_project_files("README.md")
    # dry mode test
    result = runner.invoke(cli.cli, ['init', "-vv", "--dry"])
    assert result.exit_code == 0
    assert not os.path.exists("bumpver.toml")

    # non dry mode
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    assert os.path.exists("bumpver.toml")
    with pl.Path("bumpver.toml").open(mode="r", encoding="utf-8") as fobj:
        cfg_content = fobj.read()

    base_str = config.DEFAULT_BUMPVER_TOML_BASE_TMPL.format(
        initial_version=config._initial_version(),
        default_tag_scope=config.DEFAULT_TAG_SCOPE.value,
    )
    assert base_str                          in cfg_content
    assert config.DEFAULT_TOML_README_MD_STR in cfg_content

    result = runner.invoke(cli.cli, ['show', "-vv"])
    _debug_records(caplog)
    assert result.exit_code == 0
    assert f"Current Version: {config._initial_version()}\n" in result.output
    assert f"PEP440         : {config._initial_version_pep440()}\n" in result.output

    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 1

    # check logging
    assert len(caplog.records) == 1
    log = caplog.records[0]
    assert log.levelname == 'ERROR'
    assert "Configuration already initialized" in log.message


def test_novcs_setupcfg_init(runner):
    _add_project_files("README.md", "setup.cfg")

    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    with pl.Path("setup.cfg").open(mode="r", encoding="utf-8") as fobj:
        cfg_content = fobj.read()

    base_str = config.DEFAULT_CONFIGPARSER_BASE_TMPL.format(
        initial_version=config._initial_version(),
        default_tag_scope=config.DEFAULT_TAG_SCOPE.value,
    )
    assert base_str                                  in cfg_content
    assert config.DEFAULT_CONFIGPARSER_README_MD_STR in cfg_content

    result = runner.invoke(cli.cli, ['show', "-vv"])
    assert result.exit_code == 0
    assert f"Current Version: {config._initial_version()}\n" in result.output
    assert f"PEP440         : {config._initial_version_pep440()}\n" in result.output


def test_novcs_multi_cfg(runner):
    _add_project_files("README.md", "setup.cfg")

    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    # NOTE (mb 2020-11-16): Even though a pyproject.toml exists, it shouldn't
    #   be used because the setup.cfg has the [bumpver] config section and
    #   should have higher priorty.
    with pl.Path("pyproject.toml").open(mode="wt", encoding="utf-8") as fobj:
        fobj.write("")

    result = runner.invoke(cli.cli, ['show', "-vv"])
    assert result.exit_code == 0
    assert f"Current Version: {config._initial_version()}\n" in result.output
    assert f"PEP440         : {config._initial_version_pep440()}\n" in result.output


def test_novcs_pyproject_init(runner, caplog):
    _add_project_files("README.md", "pyproject.toml")
    result = runner.invoke(cli.cli, ['init', "-vv"])
    _debug_records(caplog)
    assert result.exit_code == 0

    with pl.Path("pyproject.toml").open(mode="r", encoding="utf-8") as fobj:
        cfg_content = fobj.read()

    base_str = config.DEFAULT_PYPROJECT_TOML_BASE_TMPL.format(
        initial_version=config._initial_version(),
        default_tag_scope=config.DEFAULT_TAG_SCOPE.value,
    )
    assert base_str                          in cfg_content
    assert config.DEFAULT_TOML_README_MD_STR in cfg_content

    result = runner.invoke(cli.cli, ['show'])
    assert result.exit_code == 0
    assert f"Current Version: {config._initial_version()}\n" in result.output
    assert f"PEP440         : {config._initial_version_pep440()}\n" in result.output


def _vcs_init(vcs, files=("README.md",)):
    assert vcs in ("git", "hg")
    assert not pl.Path(f".{vcs}").exists()
    shell(f"{vcs}", "init")
    assert pl.Path(f".{vcs}").is_dir()

    for filename in files:
        shell(f"{vcs}", "add", filename)

    shell(f"{vcs}", "commit", "-m", "initial commit")


_today = dt.datetime.utcnow().date()


DEFAULT_VERSION_PATTERNS = [
    ('"vYYYY0M.BUILD[-TAG]"'    , _today.strftime("v%Y%m.1001-alpha"), _today.strftime("%Y%m.1001a0")),
    ('"vYYYY.BUILD[-TAG]"'      , _today.strftime("v%Y.1001-alpha"), _today.strftime("%Y.1001a0")),
    ('"{pycalver}"'             , _today.strftime("v%Y%m.1001-alpha"), _today.strftime("%Y%m.1001a0")),
    ('"v{year}{build}{release}"', _today.strftime("v%Y.1001-alpha"), _today.strftime("%Y.1001a0")),
]


@pytest.mark.parametrize("version_pattern, cur_version, cur_pep440", DEFAULT_VERSION_PATTERNS)
def test_git_init(runner, version_pattern, cur_version, cur_pep440):
    _add_project_files("README.md")
    _vcs_init("git")

    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val(
        "bumpver.toml",
        version_pattern=version_pattern,
        current_version='"' + cur_version + '"',
    )

    result = runner.invoke(cli.cli, ['show'])
    assert result.exit_code == 0
    assert f"Current Version: {cur_version}\n" in result.output


@pytest.mark.parametrize("version_pattern, cur_version, cur_pep440", DEFAULT_VERSION_PATTERNS)
def test_hg_init(runner, version_pattern, cur_version, cur_pep440):
    _add_project_files("README.md")
    _vcs_init("hg")

    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val(
        "bumpver.toml",
        version_pattern=version_pattern,
        current_version='"' + cur_version + '"',
    )

    result = runner.invoke(cli.cli, ['show'])
    assert result.exit_code == 0
    assert f"Current Version: {cur_version}\n" in result.output


@pytest.mark.parametrize("version_pattern, cur_version, cur_pep440", DEFAULT_VERSION_PATTERNS)
def test_v1_git_tag_eval(runner, version_pattern, cur_version, cur_pep440):
    _add_project_files("README.md")
    _vcs_init("git")

    # This will set a version that is older than the version tag
    # we set in the vcs, which should take precedence.
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val(
        "bumpver.toml",
        version_pattern=version_pattern,
        current_version='"' + cur_version + '"',
    )

    tag_version = cur_version.replace(".1001-alpha", ".1123-beta")
    assert tag_version != cur_version

    shell("git", "tag", "--annotate", tag_version, "--message", f"bump version to {tag_version}")

    result = runner.invoke(cli.cli, ['show', "-vv"])
    assert result.exit_code == 0
    assert f"Current Version: {tag_version}\n" in result.output


@pytest.mark.parametrize("version_pattern, cur_version, cur_pep440", DEFAULT_VERSION_PATTERNS)
def test_hg_tag_eval(runner, version_pattern, cur_version, cur_pep440):
    _add_project_files("README.md")
    _vcs_init("hg")

    # This will set a version that is older than the version tag
    # we set in the vcs, which should take precedence.
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val(
        "bumpver.toml",
        version_pattern=version_pattern,
        current_version='"' + cur_version + '"',
    )

    tag_version        = cur_version.replace(".1001-alpha", ".1123-beta")
    tag_version_pep440 = tag_version[1:].split(".")[0] + ".1123b0"

    shell("hg", "tag", tag_version, "--message", f"bump version to {tag_version}")

    result = runner.invoke(cli.cli, ['show', "-vv"])
    assert result.exit_code == 0
    assert f"Current Version: {tag_version}\n" in result.output
    assert f"PEP440         : {tag_version_pep440}\n" in result.output


@pytest.mark.parametrize("version_pattern, cur_version, cur_pep440", DEFAULT_VERSION_PATTERNS)
def test_novcs_bump(runner, version_pattern, cur_version, cur_pep440):
    _add_project_files("README.md")

    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val(
        "bumpver.toml",
        version_pattern=version_pattern,
        current_version='"' + cur_version + '"',
    )

    with pl.Path("README.md").open(mode="r", encoding="utf-8") as fobj:
        content = fobj.read()

    result = runner.invoke(cli.cli, ['update', "-vv"])
    assert result.exit_code == 0

    calver = cur_version.split(".")[0]

    with pl.Path("README.md").open(encoding="utf-8") as fobj:
        content = fobj.read()
        assert calver + ".1002-alpha !\n" in content
        assert calver[1:] + ".1002a0 !]\n" in content

    result = runner.invoke(cli.cli, ['update', "-vv", "--tag", "beta"])
    assert result.exit_code == 0

    with pl.Path("README.md").open(encoding="utf-8") as fobj:
        content = fobj.read()
        assert calver + ".1003-beta !\n" in content
        assert calver[1:] + ".1003b0 !]\n" in content


@pytest.mark.parametrize("version_pattern, cur_version, cur_pep440", DEFAULT_VERSION_PATTERNS)
def test_git_bump(runner, caplog, version_pattern, cur_version, cur_pep440):
    _add_project_files("README.md")
    _vcs_init("git")

    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val(
        "bumpver.toml",
        version_pattern=version_pattern,
        current_version='"' + cur_version + '"',
    )

    shell("git", "add", "bumpver.toml")
    shell("git", "commit", "-m", "initial commit")

    result = runner.invoke(cli.cli, ['update', "-vv"])
    _debug_records(caplog)
    assert result.exit_code == 0

    calver = cur_version.split(".")[0]

    with pl.Path("README.md").open(encoding="utf-8") as fobj:
        content = fobj.read()
        assert calver + ".1002-alpha !\n" in content


@pytest.mark.parametrize("version_pattern, cur_version, cur_pep440", DEFAULT_VERSION_PATTERNS)
def test_hg_bump(runner, version_pattern, cur_version, cur_pep440):
    _add_project_files("README.md")
    _vcs_init("hg")

    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val(
        "bumpver.toml",
        version_pattern=version_pattern,
        current_version='"' + cur_version + '"',
    )

    shell("hg", "add", "bumpver.toml")
    shell("hg", "commit", "-m", "initial commit")

    result = runner.invoke(cli.cli, ['update', "-vv"])
    assert result.exit_code == 0

    calver = cur_version.split(".")[0]

    with pl.Path("README.md").open(encoding="utf-8") as fobj:
        content = fobj.read()
        assert calver + ".1002-alpha !\n" in content


def test_empty_git_bump(runner, caplog):
    shell("git", "init")
    with pl.Path("setup.cfg").open(mode="w", encoding="utf-8") as fobj:
        fobj.write("")

    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    with pl.Path("setup.cfg").open(mode="r", encoding="utf-8") as fobj:
        default_cfg_data = fobj.read()

    assert "[bumpver]\n" in default_cfg_data
    assert "\ncurrent_version = " in default_cfg_data
    assert "\n[bumpver:file_patterns]\n" in default_cfg_data
    assert "\nsetup.cfg =\n" in default_cfg_data

    result = runner.invoke(cli.cli, ['update'])

    assert any(("working directory is not clean" in r.message) for r in caplog.records)
    assert any(("setup.cfg" in r.message) for r in caplog.records)


def test_empty_hg_bump(runner, caplog):
    shell("hg", "init")
    with pl.Path("setup.cfg").open(mode="w", encoding="utf-8") as fobj:
        fobj.write("")

    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    with pl.Path("setup.cfg").open(mode="r", encoding="utf-8") as fobj:
        default_cfg_text = fobj.read()

    assert "[bumpver]\n" in default_cfg_text
    assert "\ncurrent_version = " in default_cfg_text
    assert "\n[bumpver:file_patterns]\n" in default_cfg_text
    assert "\nsetup.cfg =\n" in default_cfg_text

    result = runner.invoke(cli.cli, ['update'])

    assert any(("working directory is not clean" in r.message) for r in caplog.records)
    assert any(("setup.cfg" in r.message) for r in caplog.records)


def test_incorrect_vcs_option_no_commit_tag(runner, caplog):
    _add_project_files("README.md")
    _vcs_init("git")
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    result = runner.invoke(cli.cli, ['update', "-vv", "--no-commit", "--tag-commit"])
    assert result.exit_code == 1
    assert len(caplog.records) == 1
    assert (
        "Invalid argument: --no-commit and --tag-commit cannot be used at the same time"
        in caplog.records[0].message
    )


def test_incorrect_vcs_option_no_commit_push(runner, caplog):
    _add_project_files("README.md")
    _vcs_init("git")
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val(
        "bumpver.toml",
        tag="false",
    )

    result = runner.invoke(cli.cli, ['update', "-vv", "--no-commit", "--push"])
    assert result.exit_code == 1
    assert len(caplog.records) == 1
    assert (
        "Invalid argument: --no-commit and --push cannot be used at the same time"
        in caplog.records[0].message
    )


def test_incorrect_vcs_option_tag_push(runner, caplog):
    _add_project_files("README.md")
    _vcs_init("git")
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val(
        "bumpver.toml",
        commit="false",
        tag="false",
        push="false",
    )

    result = runner.invoke(cli.cli, ['update', "-vv", "--tag-commit"])
    assert result.exit_code == 1
    assert len(caplog.records) == 1
    expected = (
        "Invalid argument: --tag-commit requires either --commit or commit=True in your config"
    )
    assert expected in caplog.records[0].message

    result = runner.invoke(cli.cli, ['update', "-vv", "--push"])
    assert result.exit_code == 1
    assert len(caplog.records) == 2
    expected = "Invalid argument: --push requires either --commit or commit=True in your config"
    assert expected in caplog.records[1].message


@pytest.mark.parametrize("version_pattern, cur_version, cur_pep440", DEFAULT_VERSION_PATTERNS)
def test_vcs_option_no_commit(runner, caplog, version_pattern, cur_version, cur_pep440):
    _add_project_files("README.md")
    _vcs_init("git")
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val(
        "bumpver.toml",
        version_pattern=version_pattern,
        current_version=f'"{cur_version}"',
    )

    shell("git", "add", "bumpver.toml")
    shell("git", "commit", "-m", "update bumpver.toml")

    prev_commit_count = shell("git", "rev-list", "HEAD", "--count").decode("utf-8")

    result = runner.invoke(cli.cli, ['update', "-vv", "--no-commit"])
    assert result.exit_code == 0

    tags = shell("git", "tag", "--list").decode("utf-8")
    assert not tags

    commit_count = shell("git", "rev-list", "HEAD", "--count").decode("utf-8")
    assert prev_commit_count == commit_count


@pytest.mark.parametrize("version_pattern, cur_version, cur_pep440", DEFAULT_VERSION_PATTERNS)
def test_vcs_option_commit(runner, caplog, version_pattern, cur_version, cur_pep440):
    _add_project_files("README.md")
    _vcs_init("git")
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val(
        "bumpver.toml",
        version_pattern=version_pattern,
        current_version=f'"{cur_version}"',
        commit="false",
        tag="false",
        push="false",
    )

    shell("git", "add", "bumpver.toml")
    shell("git", "commit", "-m", "update bumpver.toml")

    result = runner.invoke(cli.cli, ['update', "-vv", "--commit"])
    assert result.exit_code == 0

    calver = cur_version.split(".")[0]

    tags = shell("git", "tag", "--list").decode("utf-8")
    assert not tags

    last_commit = shell("git", "log", "-1").decode("utf-8")
    expected    = f"bump version {calver}.1001-alpha -> {calver}.1002-alpha"
    assert expected in last_commit


@pytest.mark.parametrize("version_pattern, cur_version, cur_pep440", DEFAULT_VERSION_PATTERNS)
def test_vcs_option_no_tag(runner, caplog, version_pattern, cur_version, cur_pep440):
    _add_project_files("README.md")
    _vcs_init("git")
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val(
        "bumpver.toml",
        version_pattern=version_pattern,
        current_version=f'"{cur_version}"',
    )

    shell("git", "add", "bumpver.toml")
    shell("git", "commit", "-m", "update bumpver.toml")

    result = runner.invoke(cli.cli, ['update', "-vv", "--no-tag-commit"])
    assert result.exit_code == 0

    calver = cur_version.split(".")[0]

    tags = shell("git", "tag", "--list").decode("utf-8")
    assert not tags

    last_commit = shell("git", "log", "-1").decode("utf-8")
    expected    = f"bump version {calver}.1001-alpha -> {calver}.1002-alpha"
    assert expected in last_commit


@pytest.mark.parametrize("version_pattern, cur_version, cur_pep440", DEFAULT_VERSION_PATTERNS)
def test_vcs_option_tag(runner, caplog, version_pattern, cur_version, cur_pep440):
    _add_project_files("README.md")
    _vcs_init("git")
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val(
        "bumpver.toml",
        version_pattern=version_pattern,
        current_version=f'"{cur_version}"',
        tag="false",
    )

    shell("git", "add", "bumpver.toml")
    shell("git", "commit", "-m", "update bumpver.toml")

    result = runner.invoke(cli.cli, ['update', "-vv", "--tag-commit"])
    assert result.exit_code == 0

    calver = cur_version.split(".")[0]

    tags = shell("git", "tag", "--list").decode("utf-8")
    assert f"{calver}.1002-alpha" in tags

    last_commit = shell("git", "log", "-1").decode("utf-8")
    expected    = f"bump version {calver}.1001-alpha -> {calver}.1002-alpha"
    assert expected in last_commit


SETUP_CFG_SEMVER_FIXTURE = """
[metadata]
license_file = LICENSE

[bdist_wheel]
universal = 1

[bumpver]
current_version = "0.1.0"
version_pattern = "{semver}"

[bumpver:file_patterns]
setup.cfg =
    current_version = "{version}"
"""


DEFAULT_SEMVER_PATTERNS = [
    '"{semver}"',
    '"MAJOR.MINOR.PATCH"',
]


@pytest.mark.parametrize("version_pattern", DEFAULT_SEMVER_PATTERNS)
def test_update_semver_warning(runner, caplog, version_pattern):
    _add_project_files("README.md")

    with pl.Path("setup.cfg").open(mode="w", encoding="utf-8") as fobj:
        fobj.write(SETUP_CFG_SEMVER_FIXTURE)

    _update_config_val("setup.cfg", version_pattern=version_pattern)

    _vcs_init("hg", files=["README.md", "setup.cfg"])

    result = runner.invoke(cli.cli, ['update', "-vv", "-n", "--dry"])
    assert result.exit_code == 1

    assert any("version did not change" in r.message for r in caplog.records)
    assert any("[--major/--minor/--patch] required" in r.message for r in caplog.records)

    result = runner.invoke(cli.cli, ['update', "-vv", "-n", "--dry", "--patch"])
    assert result.exit_code == 0


@pytest.mark.parametrize("version_pattern", DEFAULT_SEMVER_PATTERNS)
def test_update_semver_diff(runner, caplog, version_pattern):
    _add_project_files("README.md")

    with pl.Path("setup.cfg").open(mode="w", encoding="utf-8") as fobj:
        fobj.write(SETUP_CFG_SEMVER_FIXTURE)

    _update_config_val("setup.cfg", version_pattern=version_pattern)

    _vcs_init("hg", files=["README.md", "setup.cfg"])

    cases = [("--major", "1.0.0"), ("--minor", "0.2.0"), ("--patch", "0.1.1")]

    for flag, expected in cases:
        result = runner.invoke(cli.cli, ['update', "-vv", "-n", "--dry", flag])
        assert result.exit_code == 0
        assert len(caplog.records) == 0

        out_lines = set(result.output.splitlines())

        assert "+++ setup.cfg" in out_lines
        assert "-current_version = \"0.1.0\"" in out_lines
        assert f"+current_version = \"{expected}\"" in out_lines


def test_update_set_version(runner, caplog):
    _add_project_files("README.md", "setup.cfg")

    with pl.Path("setup.cfg").open(mode="w", encoding="utf-8") as fobj:
        fobj.write(SETUP_CFG_SEMVER_FIXTURE)

    _update_config_val(
        "setup.cfg",
        version_pattern='"vYYYY.BUILD[-TAG]"',
        current_version='"v2020.1001-alpha"',
    )

    result = runner.invoke(
        cli.cli, ['update', "-vv", "-n", "--dry", "--set-version", "v2121.1234-beta"]
    )
    assert result.exit_code == 0
    assert len(caplog.records) == 0

    out_lines = set(result.output.splitlines())
    assert '-current_version = "v2020.1001-alpha"' in out_lines
    assert '+current_version = "v2121.1234-beta"' in out_lines

    result = runner.invoke(
        cli.cli, ['update', "-vv", "-n", "--dry", "--set-version", "2020.1234-invalid"]
    )
    assert result.exit_code == 1
    assert len(caplog.records) > 0

    log_msg = caplog.records[0].message
    assert "Invalid version '2020.1234-invalid' for pattern 'vYYYY.BUILD[-TAG]'" in log_msg
    log_msg = caplog.records[1].message
    assert "Invalid argument --set-version='2020.1234-invalid'" in log_msg


@pytest.mark.parametrize("version_pattern, cur_version, cur_pep440", DEFAULT_VERSION_PATTERNS)
def test_get_diff(runner, version_pattern, cur_version, cur_pep440):
    _add_project_files("README.md", "setup.cfg")
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    if len(cur_pep440) == 11:
        old_version = "v2017.1002-alpha"
        old_pep440  = "2017.1002a0"
    elif len(cur_pep440) == 13:
        old_version = "v201707.1002-alpha"
        old_pep440  = "201707.1002a0"
    else:
        assert False, len(cur_pep440)

    _update_config_val(
        "setup.cfg",
        version_pattern=version_pattern,
        current_version='"' + old_version + '"',
    )
    _, cfg = config.init()
    diff_str   = cli.get_diff(cfg, cur_version)
    diff_lines = set(diff_str.splitlines())

    assert f"-        Hello World {old_version} !" in diff_lines
    assert f"+        Hello World {cur_version} !" in diff_lines

    assert f"-        [aka. {old_pep440} !]" in diff_lines
    assert f"+        [aka. {cur_pep440} !]" in diff_lines

    assert f'-current_version = "{old_version}"' in diff_lines
    assert f'+current_version = "{cur_version}"' in diff_lines


WEEKNUM_TEST_CASES = [
    # 2020-12-26  Sat
    ("2020-12-26", "YYYY.0W", "2020.51"),
    ("2020-12-26", "YYYY.0U", "2020.51"),
    ("2020-12-26", "GGGG.0V", "2020.52"),
    # 2020-12-27  Sun
    ("2020-12-27", "YYYY.0W", "2020.51"),
    ("2020-12-27", "YYYY.0U", "2020.52"),
    ("2020-12-27", "GGGG.0V", "2020.52"),
    # 2020-12-28  Mon
    ("2020-12-28", "YYYY.0W", "2020.52"),
    ("2020-12-28", "YYYY.0U", "2020.52"),
    ("2020-12-28", "GGGG.0V", "2020.53"),
    # 2020-12-29  Tue
    ("2020-12-29", "YYYY.0W", "2020.52"),
    ("2020-12-29", "YYYY.0U", "2020.52"),
    ("2020-12-29", "GGGG.0V", "2020.53"),
    # 2020-12-30  Wed
    ("2020-12-30", "YYYY.0W", "2020.52"),
    ("2020-12-30", "YYYY.0U", "2020.52"),
    ("2020-12-30", "GGGG.0V", "2020.53"),
    # 2020-12-31  Thu
    ("2020-12-31", "YYYY.0W", "2020.52"),
    ("2020-12-31", "YYYY.0U", "2020.52"),
    ("2020-12-31", "GGGG.0V", "2020.53"),
    # 2021-01-01  Fri
    ("2021-01-01", "YYYY.0W", "2021.00"),
    ("2021-01-01", "YYYY.0U", "2021.00"),
    ("2021-01-01", "GGGG.0V", "2020.53"),
    # 2021-01-02  Sat
    ("2021-01-02", "YYYY.0W", "2021.00"),
    ("2021-01-02", "YYYY.0U", "2021.00"),
    ("2021-01-02", "GGGG.0V", "2020.53"),
    # 2021-01-03  Sun
    ("2021-01-03", "YYYY.0W", "2021.00"),
    ("2021-01-03", "YYYY.0U", "2021.01"),
    ("2021-01-03", "GGGG.0V", "2020.53"),
    # 2021-01-04  Mon
    ("2021-01-04", "YYYY.0W", "2021.01"),
    ("2021-01-04", "YYYY.0U", "2021.01"),
    ("2021-01-04", "GGGG.0V", "2021.01"),
]


@pytest.mark.parametrize("date, pattern, expected", WEEKNUM_TEST_CASES)
def test_weeknum(date, pattern, expected, runner):
    cmd    = shlex.split(f"test -vv --date {date} 2020.40 {pattern}")
    result = runner.invoke(cli.cli, cmd)
    assert result.exit_code == 0
    assert "New Version: " + expected in result.output


def test_hg_commit_message(runner, caplog):
    _add_project_files("README.md", "setup.cfg")
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    commit_message = """
    "bump from {old_version} ({old_version_pep440}) to {new_version} ({new_version_pep440})"
    """
    _update_config_val(
        "setup.cfg",
        current_version='"v2019.1001-alpha"',
        version_pattern="vYYYY.BUILD[-TAG]",
        commit_message=commit_message.strip(),
    )

    _vcs_init("hg", ["README.md", "setup.cfg"])
    assert len(caplog.records) > 0

    result = runner.invoke(cli.cli, ['update', "-vv", "--pin-date", "--tag", "beta"])
    assert result.exit_code == 0

    tags = shell("hg", "tags").decode("utf-8")
    assert "v2019.1002-beta" in tags

    commits = shell(*shlex.split("hg log -l 2")).decode("utf-8").split("\n\n")

    expected = "bump from v2019.1001-alpha (2019.1001a0) to v2019.1002-beta (2019.1002b0)"
    summary  = commits[1].split("summary:")[-1]
    assert expected in summary


def test_git_commit_message(runner, caplog):
    _add_project_files("README.md", "setup.cfg")
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    commit_message = """
    "bump: {old_version} ({old_version_pep440}) -> {new_version} ({new_version_pep440})"
    """
    _update_config_val(
        "setup.cfg",
        current_version='"v2019.1001-alpha"',
        version_pattern="vYYYY.BUILD[-TAG]",
        commit_message=commit_message.strip(),
    )

    _vcs_init("git", ["README.md", "setup.cfg"])
    assert len(caplog.records) > 0

    result = runner.invoke(cli.cli, ['update', "-vv", "--pin-date", "--tag", "beta"])
    assert result.exit_code == 0

    tags = shell("git", "tag", "--list").decode("utf-8")
    assert "v2019.1002-beta" in tags

    commits = shell(*shlex.split("git log -l 2")).decode("utf-8").split("\n\n")

    expected = "bump: v2019.1001-alpha (2019.1001a0) -> v2019.1002-beta (2019.1002b0)"
    assert expected in commits[1]


def test_cli_commit_message(runner, caplog):
    _add_project_files("README.md", "setup.cfg")
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val(
        "setup.cfg",
        current_version='"v2019.1001-alpha"',
        version_pattern="vYYYY.BUILD[-TAG]",
    )

    _vcs_init("git", ["README.md", "setup.cfg"])
    assert len(caplog.records) > 0

    cmd = [
        'update',
        "-vv",
        "--pin-date",
        "--tag",
        "beta",
        "--commit-message",
        "my custom message (OLD -> NEW)",
    ]

    result = runner.invoke(cli.cli, cmd)
    assert result.exit_code == 0

    tags = shell("git", "tag", "--list").decode("utf-8")
    assert "v2019.1002-beta" in tags

    commits = shell(*shlex.split("git log -l 2")).decode("utf-8").split("\n\n")

    expected = "my custom message (v2019.1001-alpha -> v2019.1002-beta)"
    assert expected in commits[1]


def test_hg_tag_message(runner, caplog):
    _add_project_files("README.md", "setup.cfg")
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    tag_message = """
    "bump: {old_version} ({old_version_pep440}) -> {new_version} ({new_version_pep440})"
    """

    _update_config_val(
        "setup.cfg",
        current_version='"v2019.1001-alpha"',
        version_pattern="vYYYY.BUILD[-TAG]",
        tag_message=tag_message.strip(),
    )

    _vcs_init("hg", ["README.md", "setup.cfg"])
    assert len(caplog.records) > 0

    cmd = [
        'update',
        "-vv",
        "--pin-date",
        "--tag",
        "beta",
    ]

    result = runner.invoke(cli.cli, cmd)
    assert result.exit_code == 0

    tags = shell("hg", "tags").decode("utf-8")

    expected_tag = "v2019.1002-beta"
    assert expected_tag in tags

    latest_changeset = shell("hg", "log", "-l", "1").decode("utf-8")

    expected_message = "bump: v2019.1001-alpha (2019.1001a0) -> v2019.1002-beta (2019.1002b0)"
    assert expected_message in latest_changeset.split("summery:", 1)[-1]


def test_git_tag_message_annotated(runner, caplog):
    _add_project_files("README.md", "setup.cfg")
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    tag_message = """
    "bump: {old_version} ({old_version_pep440}) -> {new_version} ({new_version_pep440})"
    """

    _update_config_val(
        "setup.cfg",
        current_version='"v2019.1001-alpha"',
        version_pattern="vYYYY.BUILD[-TAG]",
        tag_message=tag_message.strip(),
    )

    _vcs_init("git", ["README.md", "setup.cfg"])
    assert len(caplog.records) > 0

    cmd = [
        'update',
        "-vv",
        "--pin-date",
        "--tag",
        "beta",
    ]

    result = runner.invoke(cli.cli, cmd)
    assert result.exit_code == 0

    tags = shell("git", "tag", "--list").decode("utf-8")

    expected_tag = "v2019.1002-beta"
    assert expected_tag in tags

    objecttype = (
        shell("git", "tag", "--list", expected_tag, "--format=%(objecttype)")
        .decode("utf-8")
        .strip()
    )

    expected_type = "tag"
    assert objecttype == expected_type

    message = (
        shell("git", "tag", "--list", expected_tag, "--format=%(contents:subject)")
        .decode("utf-8")
        .strip()
    )

    expected_message = "bump: v2019.1001-alpha (2019.1001a0) -> v2019.1002-beta (2019.1002b0)"
    assert expected_message == message


def test_git_tag_message_lightweight(runner, caplog):
    _add_project_files("README.md", "setup.cfg")
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val(
        "setup.cfg",
        current_version='"v2019.1001-alpha"',
        version_pattern="vYYYY.BUILD[-TAG]",
        tag_message='""',
    )

    _vcs_init("git", ["README.md", "setup.cfg"])
    assert len(caplog.records) > 0

    cmd = [
        'update',
        "-vv",
        "--pin-date",
        "--tag",
        "beta",
    ]

    result = runner.invoke(cli.cli, cmd)
    assert result.exit_code == 0

    tags = shell("git", "tag", "--list").decode("utf-8")

    expected_tag = "v2019.1002-beta"
    assert expected_tag in tags

    objecttype = (
        shell("git", "tag", "--list", expected_tag, "--format=%(objecttype)")
        .decode("utf-8")
        .strip()
    )

    expected_type = "commit"
    assert objecttype == expected_type


def test_cli_tag_message(runner, caplog):
    _add_project_files("README.md", "setup.cfg")
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val(
        "setup.cfg",
        current_version='"v2019.1001-alpha"',
        version_pattern="vYYYY.BUILD[-TAG]",
    )

    _vcs_init("git", ["README.md", "setup.cfg"])
    assert len(caplog.records) > 0

    cmd = [
        'update',
        "-vv",
        "--pin-date",
        "--tag",
        "beta",
        "--tag-message",
        "my custom message (OLD -> NEW)",
    ]

    result = runner.invoke(cli.cli, cmd)
    assert result.exit_code == 0

    tags = shell("git", "tag", "--list").decode("utf-8")

    expected_tag = "v2019.1002-beta"
    assert expected_tag in tags

    message = (
        shell("git", "tag", "--list", expected_tag, "--format=%(contents:subject)")
        .decode("utf-8")
        .strip()
    )

    expected_message = "my custom message (v2019.1001-alpha -> v2019.1002-beta)"
    assert expected_message == message


def test_grep(runner):
    _add_project_files("README.md")

    search_re = r"^\s+2:\s+Hello World v2017\.1002-alpha !"

    cmd1    = r'grep "vYYYY.BUILD[-TAG]" README.md'
    result1 = runner.invoke(cli.cli, shlex.split(cmd1))
    assert result1.exit_code == 0
    assert re.search(search_re, result1.output, flags=re.MULTILINE)

    cmd2    = r'grep --version-pattern "vYYYY.BUILD[-TAG]" "{version}" README.md'
    result2 = runner.invoke(cli.cli, shlex.split(cmd2))
    assert result2.exit_code == 0
    assert re.search(search_re, result2.output, flags=re.MULTILINE)

    assert result1.output == result2.output

    search_re = r"^\s+3:\s+\[aka\. 2017\.1002a0 \!\]"

    cmd3    = r'grep "\[aka. YYYY.BLD[PYTAGNUM] \!\]"  README.md'
    result3 = runner.invoke(cli.cli, shlex.split(cmd3))
    assert result3.exit_code == 0
    assert re.search(search_re, result3.output, flags=re.MULTILINE)

    cmd4    = r'grep --version-pattern "vYYYY.BUILD[-TAG]" "\[aka. {pep440_version} \!\]" README.md'
    result4 = runner.invoke(cli.cli, shlex.split(cmd4))
    assert result4.exit_code == 0
    assert re.search(search_re, result4.output, flags=re.MULTILINE)

    assert result3.output == result4.output


SETUP_CFG_MULTIMATCH_FILE_PATTERNS_FIXTURE_V1 = r"""
[pycalver]
current_version = "v201701.1002-alpha"
version_pattern = "{pycalver}"

[pycalver:file_patterns]
setup.cfg =
    current_version = "{version}"
README.md =
    Hello World {version} !
README.* =
    [aka. {pep440_version} !]
"""


SETUP_CFG_MULTIMATCH_FILE_PATTERNS_FIXTURE_V2 = r"""
[bumpver]
current_version = "v201701.1002-alpha"
version_pattern = "vYYYY0M.BUILD[-TAG][NUM]"

[bumpver:file_patterns]
setup.cfg =
    current_version = "{version}"
README.md =
    Hello World {version} !
README.* =
    \[aka. {pep440_version} !\]
"""


@pytest.mark.parametrize(
    "config_text",
    [SETUP_CFG_MULTIMATCH_FILE_PATTERNS_FIXTURE_V1, SETUP_CFG_MULTIMATCH_FILE_PATTERNS_FIXTURE_V2],
)
def test_multimatch_file_patterns(config_text, runner):
    _add_project_files("README.md")
    with pl.Path("setup.cfg").open(mode="w", encoding="utf-8") as fobj:
        fobj.write(config_text)

    with pl.Path("README.md").open(mode="r", encoding="utf-8") as fobj:
        content = fobj.read()

    assert content.count("Hello World v201707.1002-alpha !") == 1
    assert content.count("Hello World v202011.1003-beta !") == 0
    assert content.count("[aka. 201707.1002a0 !]") == 1
    assert content.count("[aka. 202011.1003b0 !]") == 0

    result = runner.invoke(cli.cli, ['update', '--tag', 'beta', '--date', "2020-11-22"])
    assert result.exit_code == 0

    with pl.Path("README.md").open(mode="r", encoding="utf-8") as fobj:
        content = fobj.read()

    assert content.count("Hello World v201707.1002-alpha !") == 0
    assert content.count("Hello World v202011.1003-beta !") == 1
    assert content.count("[aka. 201707.1002a0 !]") == 0
    assert content.count("[aka. 202011.1003b0 !]") == 1


def _kwargs(year, month, minor=False):
    return {'maybe_date': dt.date(year, month, 1), 'minor': minor}


ROLLOVER_TEST_CASES = [
    # v1 cases
    ["{year}.{month}.{MINOR}", "2020.10.3", "2020.10.4", _kwargs(2020, 10, True)],
    ["{year}.{month}.{MINOR}", "2020.10.3", None, _kwargs(2020, 10, False)],
    ["{year}.{month}.{MINOR}", "2020.10.3", "2020.11.4", _kwargs(2020, 11, True)],
    ["{year}.{month}.{MINOR}", "2020.10.3", "2020.11.3", _kwargs(2020, 11, False)],
    # v2 cases
    ["YYYY.MM.MINOR", "2020.10.3", "2020.10.4", _kwargs(2020, 10, True)],
    ["YYYY.MM.MINOR", "2020.10.3", None, _kwargs(2020, 10, False)],
    ["YYYY.MM.MINOR", "2020.10.3", "2020.11.0", _kwargs(2020, 11, True)],
    ["YYYY.MM.MINOR", "2020.10.3", "2020.11.0", _kwargs(2020, 11, False)],
    ["YYYY.MM[.MINOR]", "2020.10.3", "2020.10.4", _kwargs(2020, 10, True)],
    ["YYYY.MM[.MINOR]", "2020.10.3", "2020.11", _kwargs(2020, 11, False)],
    ["YYYY.MM.MINOR", "2020.10.3", "2021.10.0", _kwargs(2021, 10, False)],
    # incr0/incr1 part
    ["YYYY.MM.INC0", "2020.10.3", "2020.10.4", _kwargs(2020, 10)],
    ["YYYY.MM.INC0", "2020.10.3", "2020.11.0", _kwargs(2020, 11)],
    ["YYYY.MM.INC0", "2020.10.3", "2021.10.0", _kwargs(2021, 10)],
    ["YYYY.MM.INC1", "2020.10.3", "2020.10.4", _kwargs(2020, 10)],
    ["YYYY.MM.INC1", "2020.10.3", "2020.11.1", _kwargs(2020, 11)],
    ["YYYY.MM.INC1", "2020.10.3", "2021.10.1", _kwargs(2021, 10)],
]


@pytest.mark.parametrize("version_pattern, old_version, expected, kwargs", ROLLOVER_TEST_CASES)
def test_rollover(version_pattern, old_version, expected, kwargs):
    new_version = cli.incr_dispatch(old_version, raw_pattern=version_pattern, **kwargs)
    if new_version is None:
        assert expected is None
    else:
        assert new_version == expected


VCS_VERSION_TAG_CASES = [
    ("git", None, "0.2.1"),
    ("git", config.TagScope.GLOBAL, "0.2.1"),
    ("git", config.TagScope.BRANCH, "0.1.10"),
    ("git", config.TagScope.DEFAULT, "0.2.1"),
    ("hg", None, "0.2.1"),
    ("hg", config.TagScope.GLOBAL, "0.2.1"),
    ("hg", config.TagScope.BRANCH, "0.1.10"),
    ("hg", config.TagScope.DEFAULT, "0.2.1"),
]


@pytest.mark.parametrize("vcs_name, tag_scope, expected_version", VCS_VERSION_TAG_CASES)
def test_get_latest_vcs_version_tag(runner, vcs_name, tag_scope, expected_version):
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val("bumpver.toml", push="false")
    _update_config_val("bumpver.toml", current_version='"0.1.8"')
    _update_config_val("bumpver.toml", version_pattern='"MAJOR.MINOR.PATCH[PYTAGNUM]"')

    if tag_scope is not None:
        _update_config_val("bumpver.toml", tag_scope=f'"{tag_scope.value}"')

    _vcs_init(vcs_name, files=["bumpver.toml"])

    result = runner.invoke(cli.cli, ['update', "--patch"])
    assert result.exit_code == 0

    _, cfg = config.init()
    latest_version = cli.get_latest_vcs_version_tag(cfg, fetch=False)
    assert latest_version == "0.1.9"

    if vcs_name == 'git':
        shell("git", "branch", "dev")
        shell("git", "checkout", "dev")
    else:
        shell("hg", "branch", "dev")

    result = runner.invoke(cli.cli, ['update', "--minor", "--tag=beta"])
    assert result.exit_code == 0

    _, cfg = config.init()
    latest_version = cli.get_latest_vcs_version_tag(cfg, fetch=False)
    assert latest_version == "0.2.0b0"

    if vcs_name == 'git':
        shell("git", "checkout", "master")
    else:
        shell("hg", "update", "default")

    result = runner.invoke(cli.cli, ['update', "--patch", "--tag=final"])
    assert result.exit_code == 0

    _, cfg = config.init()
    latest_version = cli.get_latest_vcs_version_tag(cfg, fetch=False)
    assert latest_version == expected_version


@pytest.mark.parametrize("vcs_name", ['git', 'hg'])
def test_ignore_vcs_tag(runner, monkeypatch, vcs_name):
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val("bumpver.toml", push="false")
    _update_config_val("bumpver.toml", current_version='"0.1.8"')
    _update_config_val("bumpver.toml", version_pattern='"MAJOR.MINOR.PATCH"')

    _vcs_init(vcs_name, files=["bumpver.toml"])
    _, cfg = config.init()

    # mock latest vcs tag 0.2.0 but cfg.current_version is 0.1.8
    monkeypatch.setattr(cli, "get_latest_vcs_version_tag", lambda cfg, fetch: "0.2.0")
    assert cfg.current_version == "0.1.8"
    assert cli.get_latest_vcs_version_tag(cfg, fetch=False) == "0.2.0"

    result = runner.invoke(cli.cli, ['update', "--set-version", "0.1.9"])
    assert result.exit_code == 1

    result = runner.invoke(
        cli.cli, ['update', "-vv", "--ignore-vcs-tag", "--dry", "--set-version", "0.1.9"]
    )
    assert result.exit_code == 0

    out_lines = set(result.output.splitlines())
    assert '-current_version = "0.1.8"' in out_lines
    assert '+current_version = "0.1.9"' in out_lines

    latest_version = cli.get_latest_vcs_version_tag(cfg, fetch=False)
    assert latest_version == "0.2.0"


@pytest.mark.parametrize("vcs_name", ['git', 'hg'])
def test_git_tag_scope_branch_version_conflict(runner, caplog, vcs_name):
    result = runner.invoke(cli.cli, ['init', "-vv"])
    assert result.exit_code == 0

    _update_config_val("bumpver.toml", push="false")
    _update_config_val("bumpver.toml", current_version='"0.1.8"')
    _update_config_val("bumpver.toml", version_pattern='"MAJOR.MINOR.PATCH[PYTAGNUM]"')
    _update_config_val("bumpver.toml", tag_scope=f'"{config.TagScope.BRANCH.value}"')

    _vcs_init(vcs_name, files=["bumpver.toml"])

    result = runner.invoke(cli.cli, ['update', "--patch"])
    assert result.exit_code == 0

    _, cfg = config.init()
    latest_version = cli.get_latest_vcs_version_tag(cfg, fetch=False)
    assert latest_version == "0.1.9"

    if vcs_name == 'git':
        shell("git", "branch", "dev")
        shell("git", "checkout", "dev")
    else:
        shell("hg", "branch", "dev")

    result = runner.invoke(cli.cli, ['update', "--minor", "--tag=beta"])
    assert result.exit_code == 0

    _, cfg = config.init()
    latest_version = cli.get_latest_vcs_version_tag(cfg, fetch=False)
    assert latest_version == "0.2.0b0"

    if vcs_name == 'git':
        shell("git", "checkout", "master")
    else:
        shell("hg", "update", "default")

    result = runner.invoke(cli.cli, ['update', "--minor", "--tag=beta"])
    assert result.exit_code == 1

    error_message = "Invariant violated: New version must be unique accross all branches"
    assert any(error_message in r.message for r in caplog.records)


def test_commit_hook_execution(runner, caplog):
    result = runner.invoke(cli.cli, ['init'])
    assert result.exit_code == 0

    _update_config_val("bumpver.toml", push="false")
    _update_config_val("bumpver.toml", current_version='"0.1.0"')
    _update_config_val("bumpver.toml", version_pattern='"MAJOR.MINOR.PATCH"')

    _vcs_init("git", files=["bumpver.toml"])

    hooks_path = util.FIXTURES_DIR / "hooks"

    cmd = [
        'update',
        '--minor',
        '--pre-commit-hook',
        f'{str(hooks_path / "pre_commit_hook.py")}',
        '--post-commit-hook',
        f'{str(hooks_path / "post_commit_hook.py")}',
    ]

    caplog.set_level(logging.INFO)
    result = runner.invoke(cli.cli, cmd)
    assert result.exit_code == 0

    pre_commit_msg = "I'm a pre commit hook. The old version was: 0.1.0"
    assert any(pre_commit_msg in r.message for r in caplog.records)

    post_commit_msg = "I'm a post commit hook. The new version is: 0.2.0"
    assert any(post_commit_msg in r.message for r in caplog.records)


def test_commit_hook_execution_fail(runner, caplog):
    result = runner.invoke(cli.cli, ['init'])
    assert result.exit_code == 0

    _update_config_val("bumpver.toml", push="false")

    _vcs_init("git", files=["bumpver.toml"])

    hooks_path = util.FIXTURES_DIR / "hooks"

    cmd = [
        'update',
        '--pre-commit-hook',
        f'{str(hooks_path / "pre_commit_hook_fail.py")}',
    ]

    result = runner.invoke(cli.cli, cmd)
    assert result.exit_code == 1
    assert len(caplog.records) > 0

    log_msg = caplog.records[0].message
    assert "Script exited with an error. Stopping" in log_msg



# This file is part of the bumpver project
# https://github.com/mbarkhau/bumpver
#
# Copyright (c) 2018-2023 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
import typing as typ
import datetime as dt

MaybeInt = typ.Optional[int]


def parse_version(version: str) -> typ.Any:
    # pylint: disable=import-outside-toplevel; lazy import to speed up --help

    try:
        import pkg_resources

        return pkg_resources.parse_version(version)
    except (ImportError, ValueError):
        import looseversion

        return looseversion.LooseVersion(version)


class V1CalendarInfo(typ.NamedTuple):
    """Container for calendar components of version strings."""

    year    : MaybeInt
    quarter : MaybeInt
    month   : MaybeInt
    dom     : MaybeInt
    doy     : MaybeInt
    iso_week: MaybeInt
    us_week : MaybeInt


class V1VersionInfo(typ.NamedTuple):
    """Container for parsed version string."""

    year    : MaybeInt
    quarter : MaybeInt
    month   : MaybeInt
    dom     : MaybeInt
    doy     : MaybeInt
    iso_week: MaybeInt
    us_week : MaybeInt
    major   : int
    minor   : int
    patch   : int
    bid     : str
    tag     : str


class V2CalendarInfo(typ.NamedTuple):
    """Container for calendar components of version strings."""

    year_y : MaybeInt
    year_g : MaybeInt
    quarter: MaybeInt
    month  : MaybeInt
    dom    : MaybeInt
    doy    : MaybeInt
    week_w : MaybeInt
    week_u : MaybeInt
    week_v : MaybeInt


class V2VersionInfo(typ.NamedTuple):
    """Container for parsed version string."""

    year_y : MaybeInt
    year_g : MaybeInt
    quarter: MaybeInt
    month  : MaybeInt
    dom    : MaybeInt
    doy    : MaybeInt
    week_w : MaybeInt
    week_u : MaybeInt
    week_v : MaybeInt
    major  : int
    minor  : int
    patch  : int
    bid    : str
    tag    : str
    pytag  : str
    githash: str
    hexhash: str
    num    : int
    inc0   : int
    inc1   : int


# The test suite may replace this.
TODAY = dt.datetime.utcnow().date()


TAG_BY_PEP440_TAG = {
    'a'   : 'alpha',
    'b'   : 'beta',
    ''    : 'final',
    'rc'  : 'rc',
    'dev' : 'dev',
    'post': 'post',
}


PEP440_TAG_BY_TAG = {
    'a'      : 'a',
    'b'      : 'b',
    'dev'    : 'dev',
    'alpha'  : 'a',
    'beta'   : 'b',
    'preview': 'rc',
    'pre'    : 'rc',
    'rc'     : 'rc',
    'c'      : 'rc',
    'final'  : '',
    'post'   : 'post',
    'r'      : 'post',
    'rev'    : 'post',
}

assert set(TAG_BY_PEP440_TAG.keys()) == set(PEP440_TAG_BY_TAG.values())
assert set(TAG_BY_PEP440_TAG.values()) < set(PEP440_TAG_BY_TAG.keys())


PART_ZERO_VALUES = {
    'MAJOR': "0",
    'MINOR': "0",
    'PATCH': "0",
    'TAG'  : "final",
    'PYTAG': "",
    'NUM'  : "0",
    'INC0' : "0",
}


V2_FIELD_INITIAL_VALUES = {
    'major': "0",
    'minor': "0",
    'patch': "0",
    'num'  : "0",
    'inc0' : "0",
    'inc1' : "1",
}


def is_zero_val(part: str, part_value: str) -> bool:
    return part in PART_ZERO_VALUES and part_value == PART_ZERO_VALUES[part]


class PatternError(Exception):
    pass


def date_from_doy(year: int, doy: int) -> dt.date:
    """Parse date from year and day of year (1 indexed).

    >>> cases = [
    ...     (2016, 1), (2016, 31), (2016, 31 + 1), (2016, 31 + 29), (2016, 31 + 30),
    ...     (2017, 1), (2017, 31), (2017, 31 + 1), (2017, 31 + 28), (2017, 31 + 29),
    ... ]
    >>> dates = [date_from_doy(year, month) for year, month in cases]
    >>> assert [(d.month, d.day) for d in dates] == [
    ...     (1, 1), (1, 31), (2, 1), (2, 29), (3, 1),
    ...     (1, 1), (1, 31), (2, 1), (2, 28), (3, 1),
    ... ]
    """
    return dt.date(year, 1, 1) + dt.timedelta(days=doy - 1)


def quarter_from_month(month: int) -> int:
    """Calculate quarter (1 indexed) from month (1 indexed).

    >>> [quarter_from_month(month) for month in range(1, 13)]
    [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4]
    """
    return ((month - 1) // 3) + 1


def to_pep440(version: str) -> str:
    """Derive pep440 compliant version string from PyCalVer version string.

    >>> to_pep440("v201811.0007-beta")
    '201811.7b0'
    """
    return str(parse_version(version))



#!/usr/bin/env python
# This file is part of the bumpver project
# https://github.com/mbarkhau/bumpver
#
# Copyright (c) 2018-2023 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
"""cli module for BumpVer."""
import io
import re
import sys
import typing as typ
import logging
import datetime as dt
import subprocess as sp

import click
import colorama

from . import vcs
from . import config
from . import rewrite
from . import version
from . import patterns
from . import regexfmt
from . import v1rewrite
from . import v1version
from . import v2rewrite
from . import v2version
from . import v1patterns
from . import v2patterns

try:
    import pretty_traceback

    pretty_traceback.install()
except ImportError:
    pass  # no need to fail because of missing dev dependency


click.disable_unicode_literals_warning = True

logger = logging.getLogger("bumpver.cli")


_VERBOSE = 0


# pylint:disable=too-many-arguments; such is the cli
# pylint:disable=too-many-locals; such is the cli


def _configure_logging(verbose: int = 0) -> None:
    # pylint:disable=global-statement; global flag is global.
    global _VERBOSE
    _VERBOSE = verbose

    if verbose >= 2:
        log_format = "%(asctime)s.%(msecs)03d %(levelname)-7s %(name)-17s - %(message)s"
        log_level  = logging.DEBUG
    elif verbose == 1:
        log_format = "%(levelname)-7s - %(message)s"
        log_level  = logging.INFO
    else:
        log_format = "%(levelname)-7s - %(message)s"
        log_level  = logging.INFO

    logging.basicConfig(level=log_level, format=log_format, datefmt="%Y-%m-%dT%H:%M:%S")
    logger.debug("Logging configured.")


VALID_RELEASE_TAG_VALUES = ("alpha", "beta", "dev", "rc", "post", "final")


_current_date = dt.date.today().isoformat()


def _validate_date(date: typ.Optional[str], pin_date: bool) -> typ.Optional[dt.date]:
    if date and pin_date:
        logger.error(f"Can only use either --pin-date or --date='{date}', not both.")
        sys.exit(1)

    if date is None:
        return None

    try:
        dt_val = dt.datetime.strptime(date, "%Y-%m-%d")
        return dt_val.date()
    except ValueError:
        logger.error(
            f"Invalid parameter --date='{date}', must match format YYYY-0M-0D.", exc_info=True
        )
        sys.exit(1)


def _validate_release_tag(tag: typ.Optional[str]) -> None:
    if tag is None:
        return

    if tag in VALID_RELEASE_TAG_VALUES:
        return

    logger.error(f"Invalid argument --tag={tag}")
    logger.error(f"Valid arguments are: {', '.join(VALID_RELEASE_TAG_VALUES)}")
    sys.exit(1)


def _validate_flags(
    raw_pattern: str,
    major      : bool,
    minor      : bool,
    patch      : bool,
) -> None:
    if "{" in raw_pattern and "}" in raw_pattern:
        # only validate for new style patterns
        return

    valid = True
    if major and "MAJOR" not in raw_pattern:
        logger.error(f"Flag --major is not applicable to pattern '{raw_pattern}'")
        valid = False
    if minor and "MINOR" not in raw_pattern:
        logger.error(f"Flag --minor is not applicable to pattern '{raw_pattern}'")
        valid = False
    if patch and "PATCH" not in raw_pattern:
        logger.error(f"Flag --patch is not applicable to pattern '{raw_pattern}'")
        valid = False

    if not valid:
        sys.exit(1)


def _log_no_change(subcmd: str, version_pattern: str) -> None:
    is_semver = "{semver}" in version_pattern or (
        "MAJOR" in version_pattern and "MAJOR" in version_pattern and "PATCH" in version_pattern
    )
    if is_semver:
        logger.warning(f"bumpver {subcmd} [--major/--minor/--patch] required for use with SemVer.")
    else:
        available_flags = [
            "--" + part.lower() for part in ['MAJOR', 'MINOR', 'PATCH'] if part in version_pattern
        ]
        if available_flags:
            available_flags_str = "/".join(available_flags)
            logger.info(f"Perhaps try: bumpver {subcmd} {available_flags_str} ")


def _get_normalized_pattern(raw_pattern: str, version_pattern: typ.Optional[str]) -> str:
    is_version_pattern_required = "{version}" in raw_pattern or "{pep440_version}" in raw_pattern

    if is_version_pattern_required and version_pattern is None:
        logger.error(
            "Argument --version-pattern=<PATTERN> is required"
            " for placeholders: {version}/{pep440_version}."
        )
        sys.exit(1)
    elif version_pattern is None:
        _version_pattern = "INVALID"  # pacify mypy, it's not referenced in raw_pattern
    else:
        _version_pattern = version_pattern

    if is_version_pattern_required:
        return v2patterns.normalize_pattern(_version_pattern, raw_pattern)
    else:
        return raw_pattern


verbose_option = click.option(
    "-v",
    "--verbose",
    count=True,
    help="Control log level. -vv for debug level.",
)


dry_option = click.option(
    "-d",
    "--dry",
    default=False,
    is_flag=True,
    help="Display diff of changes, don't rewrite files.",
)


fetch_option = click.option(
    "-f/-n",
    "--fetch/--no-fetch",
    is_flag=True,
    default=True,
    help="Sync tags from remote origin.",
)


env_option = click.option(
    "-e",
    "--env",
    is_flag=True,
    default=False,
    help="Print version state for use with shell scripts: eval $(bumpver show --env)",
)


def version_options(function: typ.Callable) -> typ.Callable:
    decorators = [
        click.option("--major", is_flag=True, default=False, help="Increment MAJOR component."),
        click.option(
            "-m", "--minor", is_flag=True, default=False, help="Increment MINOR component."
        ),
        click.option(
            "-p", "--patch", is_flag=True, default=False, help="Increment PATCH component."
        ),
        click.option(
            "-t",
            "--tag",
            default=None,
            metavar="<NAME>",
            help=(
                "Override release tag of current_version. Valid options are: "
                f"{', '.join(VALID_RELEASE_TAG_VALUES)}."
            ),
        ),
        click.option(
            "--tag-num",
            is_flag=True,
            default=False,
            help="Increment release tag number (rc1, rc2, rc3..).",
        ),
        click.option(
            "--pin-increments",
            is_flag=True,
            default=False,
            help="Leave the auto-increments INC0 and INC1 unchanged.",
        ),
        click.option(
            "--pin-date", is_flag=True, default=False, help="Leave date components unchanged."
        ),
        click.option(
            "--date",
            default=None,
            metavar="<ISODATE>",
            help=f"Set explicit date in format YYYY-0M-0D (e.g. {_current_date}).",
        ),
        click.option(
            "--set-version",
            default=None,
            metavar="<VERSION>",
            help="Set version explicitly.",
        ),
    ]
    decorated = function
    for decorator in decorators:
        decorated = decorator(decorated)
    return decorated


@click.group(context_settings={'help_option_names': ["-h", "--help"]})
@click.version_option(version="2023.1127")
@verbose_option
def cli(verbose: int = 0) -> None:
    """Automatically update version strings in plaintext files."""
    if verbose:
        _configure_logging(verbose=max(_VERBOSE, verbose))


@cli.command()
@click.argument("old_version")
@click.argument("pattern")
@verbose_option
@version_options
def test(
    old_version   : str,
    pattern       : str,
    verbose       : int = 0,
    major         : bool = False,
    minor         : bool = False,
    patch         : bool = False,
    tag           : typ.Optional[str] = None,
    tag_num       : bool = False,
    pin_increments: bool = False,
    pin_date      : bool = False,
    date          : typ.Optional[str] = None,
    set_version   : typ.Optional[str] = None,
) -> None:
    """Increment a version number for demo purposes."""
    _configure_logging(verbose=max(_VERBOSE, verbose))
    _validate_release_tag(tag)

    raw_pattern = pattern  # use internal naming convention

    _validate_flags(raw_pattern, major, minor, patch)
    maybe_date = _validate_date(date, pin_date)

    if set_version is None:
        new_version = incr_dispatch(
            old_version,
            raw_pattern=raw_pattern,
            major=major,
            minor=minor,
            patch=patch,
            tag=tag,
            tag_num=tag_num,
            pin_increments=pin_increments,
            pin_date=pin_date,
            maybe_date=maybe_date,
        )
    else:
        new_version = set_version

    if new_version is None:
        _log_no_change('test', raw_pattern)
        sys.exit(1)

    if not _is_valid_version(raw_pattern, old_version, new_version):
        if set_version:
            logger.error(f"Invalid argument --set-version='{set_version}'")

        sys.exit(1)

    pep440_version = version.to_pep440(new_version)

    click.echo(f"New Version: {new_version}")
    if new_version != pep440_version:
        click.echo(f"PEP440     : {pep440_version}")


def _grep_text(pattern: patterns.Pattern, text: str, color: bool) -> typ.Iterable[str]:
    all_lines = text.splitlines()
    for match in pattern.regexp.finditer(text):
        match_start, match_end = match.span()

        line_idx   = text[:match_start].count("\n")
        line_start = text.rfind("\n", 0, match_start) + 1
        line_end   = text.find("\n", match_end, -1)
        if color:
            matched_line = (
                text[line_start:match_start]
                + colorama.Style.BRIGHT
                + text[match_start:match_end]
                + colorama.Style.RESET_ALL
                + text[match_end:line_end]
            )
        else:
            matched_line = (
                text[line_start:match_start]
                + text[match_start:match_end]
                + text[match_end:line_end]
            )

        lines_offset = max(0, line_idx - 1) + 1
        lines        = all_lines[line_idx - 1 : line_idx + 2]

        if line_idx == 0:
            lines[0] = matched_line
        else:
            lines[1] = matched_line

        prefixed_lines = [f"{lines_offset + i:>4}: {line}" for i, line in enumerate(lines)]
        yield "\n".join(prefixed_lines)


def _grep(
    raw_pattern: str,
    file_ios   : typ.Tuple[io.TextIOWrapper],
    color      : bool,
) -> None:
    pattern = v2patterns.compile_pattern(raw_pattern)

    match_count = 0
    for file_io in file_ios:
        text = file_io.read()

        match_strs = list(_grep_text(pattern, text, color))
        if len(match_strs) > 0:
            if len(file_ios) > 1:
                print(file_io.name)
            for match_str in match_strs:
                print(match_str)
            print()

        match_count += len(match_strs)

    if match_count == 0:
        logger.error(f"Pattern not found: '{raw_pattern}'")

    if match_count == 0 or _VERBOSE:
        pyexpr_regex = regexfmt.pyexpr_regex(pattern.regexp.pattern)

        print("# " + regexfmt.regex101_url(pattern.regexp.pattern))
        print(pyexpr_regex)
        print()

    if match_count == 0:
        sys.exit(1)


@cli.command()
@verbose_option
@click.option(
    "--version-pattern",
    default=None,
    metavar="<PATTERN>",
    help="Pattern to use for placeholders: {version}/{pep440_version}",
)
@click.argument("pattern")
@click.argument('files', nargs=-1, type=click.File('r'))
def grep(
    pattern        : str,
    files          : typ.Tuple[io.TextIOWrapper],
    version_pattern: typ.Optional[str] = None,
    verbose        : int = 0,
) -> None:
    """Search file(s) for a version pattern."""
    verbose = max(_VERBOSE, verbose)
    _configure_logging(verbose)

    raw_pattern        = pattern  # use internal naming convention
    normalized_pattern = _get_normalized_pattern(raw_pattern, version_pattern)

    isatty = getattr(sys.stdout, 'isatty', lambda: False)

    if isatty():
        colorama.init()
        try:
            _grep(normalized_pattern, files, color=True)
        finally:
            colorama.deinit()
    else:
        _grep(normalized_pattern, files, color=False)


@cli.command()
@verbose_option
@fetch_option
@env_option
def show(verbose: int = 0, fetch: bool = True, env: bool = False) -> None:
    """Show current version of your project."""
    _configure_logging(verbose=max(_VERBOSE, verbose))

    _, cfg = config.init(project_path=".")

    if cfg is None:
        logger.error("Could not parse configuration. Perhaps try 'bumpver init'.")
        sys.exit(1)

    cfg = _update_cfg_from_vcs(cfg, fetch)
    if env:
        version_info = v2version.parse_version_info(cfg.current_version, cfg.version_pattern)
        for key, val in version_info._asdict().items():
            click.echo(f"{key.upper()}={val if val else ''}")
        click.echo(f"CURRENT_VERSION={cfg.current_version}")
        click.echo(f"PEP440_VERSION={cfg.pep440_version}")
    else:
        click.echo(f"Current Version: {cfg.current_version}")
        click.echo(f"PEP440         : {cfg.pep440_version}")


def _colored_diff_lines(diff: str) -> typ.Iterable[str]:
    for line in diff.splitlines():
        if line.startswith("+++") or line.startswith("---"):
            yield line
        elif line.startswith("+"):
            yield "\u001b[32m" + line + "\u001b[0m"
        elif line.startswith("-"):
            yield "\u001b[31m" + line + "\u001b[0m"
        elif line.startswith("@"):
            yield "\u001b[36m" + line + "\u001b[0m"
        else:
            yield line


def _v2_get_diff(cfg: config.Config, new_version: str) -> str:
    old_vinfo = v2version.parse_version_info(cfg.current_version, cfg.version_pattern)
    new_vinfo = v2version.parse_version_info(new_version, cfg.version_pattern)
    return v2rewrite.diff(old_vinfo, new_vinfo, cfg.file_patterns)


def _v1_get_diff(cfg: config.Config, new_version: str) -> str:
    old_vinfo = v1version.parse_version_info(cfg.current_version, cfg.version_pattern)
    new_vinfo = v1version.parse_version_info(new_version, cfg.version_pattern)
    return v1rewrite.diff(old_vinfo, new_vinfo, cfg.file_patterns)


def get_diff(cfg, new_version) -> str:
    if cfg.is_new_pattern:
        return _v2_get_diff(cfg, new_version)
    else:
        return _v1_get_diff(cfg, new_version)


def _print_diff_str(diff: str) -> None:
    colored_diff = "\n".join(_colored_diff_lines(diff))
    if sys.stdout.isatty():
        click.echo(colored_diff)
    else:
        click.echo(diff)


def _print_diff(cfg: config.Config, new_version: str) -> None:
    try:
        diff = get_diff(cfg, new_version)
        _print_diff_str(diff)
    except OSError as err:
        logger.error(str(err))
        sys.exit(1)
    except rewrite.NoPatternMatch as ex:
        logger.error(str(ex))
        sys.exit(1)


def _parse_version_tags(
    all_tags: typ.List[str], version_pattern: str, is_new_pattern: bool
) -> typ.List[str]:
    version_parser = v2version if is_new_pattern else v1version
    return [tag for tag in all_tags if version_parser.is_valid(tag, version_pattern)]


def _is_valid_version(
    raw_pattern: str, old_version: str, new_version: str, unique: bool = False
) -> bool:
    is_new_pattern = "{" not in raw_pattern and "}" not in raw_pattern

    try:
        if is_new_pattern:
            v2version.parse_version_info(new_version, raw_pattern)
        else:
            v1version.parse_version_info(new_version, raw_pattern)
    except version.PatternError:
        logger.error(f"Invalid version '{new_version}' for pattern '{raw_pattern}'")
        return False

    if version.parse_version(new_version) <= version.parse_version(old_version):
        logger.error("Invariant violated: New version must be greater than old version ")
        logger.error(f"  Failed Invariant: '{new_version}' > '{old_version}'")
        logger.error("If the invariant is from vcs tags try '--ignore-vcs-tag' option.")
        return False

    if unique:
        all_tags     = vcs.get_tags(fetch=False, scope=config.TagScope.GLOBAL)
        version_tags = _parse_version_tags(all_tags, raw_pattern, is_new_pattern)

        if new_version in version_tags:
            logger.error("Invariant violated: New version must be unique accross all branches")
            return False

    return True


def incr_dispatch(
    old_version: str,
    raw_pattern: str,
    *,
    major         : bool = False,
    minor         : bool = False,
    patch         : bool = False,
    tag           : typ.Optional[str] = None,
    tag_num       : bool = False,
    pin_increments: bool = False,
    pin_date      : bool = False,
    maybe_date    : typ.Optional[dt.date] = None,
) -> typ.Optional[str]:
    v1_parts    = list(v1patterns.PART_PATTERNS) + list(v1patterns.FULL_PART_FORMATS)
    has_v1_part = any("{" + part + "}" in raw_pattern for part in v1_parts)

    if _VERBOSE:
        if has_v1_part:
            pattern = v1patterns.compile_pattern(raw_pattern)
        else:
            pattern = v2patterns.compile_pattern(raw_pattern)

        logger.info("Using pattern " + raw_pattern)
        logger.info("regex = " + regexfmt.pyexpr_regex(pattern.regexp.pattern))

    if has_v1_part:
        return v1version.incr(
            old_version,
            raw_pattern=raw_pattern,
            major=major,
            minor=minor,
            patch=patch,
            tag=tag,
            tag_num=tag_num,
            pin_date=pin_date,
            maybe_date=maybe_date,
        )
    else:
        return v2version.incr(
            old_version,
            raw_pattern=raw_pattern,
            major=major,
            minor=minor,
            patch=patch,
            tag=tag,
            tag_num=tag_num,
            pin_increments=pin_increments,
            pin_date=pin_date,
            maybe_date=maybe_date,
        )


def _update(
    cfg           : config.Config,
    new_version   : str,
    commit_message: str,
    tag_message   : str,
    allow_dirty   : bool = False,
) -> None:
    vcs_api: typ.Optional[vcs.VCSAPI] = None

    if cfg.commit:
        try:
            vcs_api = vcs.get_vcs_api()
        except OSError:
            logger.warning("Version Control System not found, skipping commit.")

    filepaths = set(cfg.file_patterns.keys())

    if vcs_api:
        vcs.assert_not_dirty(vcs_api, filepaths, allow_dirty)

    try:
        if cfg.is_new_pattern:
            new_v2_vinfo = v2version.parse_version_info(new_version, cfg.version_pattern)
            v2rewrite.rewrite_files(cfg.file_patterns, new_v2_vinfo)
        else:
            new_v1_vinfo = v1version.parse_version_info(new_version, cfg.version_pattern)
            v1rewrite.rewrite_files(cfg.file_patterns, new_v1_vinfo)
    except rewrite.NoPatternMatch as ex:
        logger.error(str(ex))
        sys.exit(1)

    if vcs_api:
        vcs.commit(cfg, vcs_api, filepaths, new_version, commit_message, tag_message)


def _try_update(
    cfg           : config.Config,
    new_version   : str,
    commit_message: str,
    tag_message   : str,
    allow_dirty   : bool = False,
) -> None:
    try:
        _update(cfg, new_version, commit_message, tag_message, allow_dirty)
    except sp.CalledProcessError as ex:
        logger.error(f"Error running subcommand: {ex.cmd}")
        if ex.stdout:
            sys.stdout.write(ex.stdout.decode('utf-8'))
        if ex.stderr:
            sys.stderr.write(ex.stderr.decode('utf-8'))
        sys.exit(1)


@cli.command()
@verbose_option
@dry_option
def init(verbose: int = 0, dry: bool = False) -> None:
    """Initialize [bumpver] configuration."""
    _configure_logging(verbose=max(_VERBOSE, verbose))

    ctx, cfg = config.init(project_path=".", cfg_missing_ok=True)

    if cfg:
        logger.error(f"Configuration already initialized in {ctx.config_rel_path}")
        sys.exit(1)

    if dry:
        click.echo(f"Exiting because of '-d/--dry'. Would have written to {ctx.config_rel_path}:")
        cfg_text: str = config.default_config(ctx)
        click.echo("\n    " + "\n    ".join(cfg_text.splitlines()))
        sys.exit(0)

    config.write_content(ctx)


def get_latest_vcs_version_tag(cfg: config.Config, fetch: bool) -> typ.Optional[str]:
    all_tags     = vcs.get_tags(fetch=fetch, scope=cfg.tag_scope)
    version_tags = _parse_version_tags(all_tags, cfg.version_pattern, cfg.is_new_pattern)

    if version_tags:
        version_tags.sort(key=version.parse_version, reverse=True)
        _debug_tags = ", ".join(version_tags[:3])
        logger.debug(f"found tags: {_debug_tags} ... ({len(version_tags)} in total)")
        return version_tags[0]
    else:
        return None


def _update_cfg_from_vcs(cfg: config.Config, fetch: bool) -> config.Config:
    latest_version_tag = get_latest_vcs_version_tag(cfg, fetch)

    if latest_version_tag is None:
        logger.debug("no vcs tags found")
        return cfg

    latest_version_pep440 = version.to_pep440(latest_version_tag)

    scope_str = f"({cfg.tag_scope.value})" if not cfg.tag_scope == config.TagScope.DEFAULT else ""
    logger.info(f"Latest version from VCS tag: {latest_version_tag} {scope_str}")

    if cfg.tag_scope == config.TagScope.DEFAULT:
        logger.info(f"Working dir version        : {cfg.current_version}")
        if version.parse_version(latest_version_tag) <= version.parse_version(cfg.current_version):
            # current_version already newer/up-to-date
            return cfg

    return cfg._replace(
        current_version=latest_version_tag,
        pep440_version=latest_version_pep440,
    )


def _parse_vcs_options(
    cfg             : config.Config,
    commit          : typ.Optional[bool] = None,
    tag_commit      : typ.Optional[bool] = None,
    push            : typ.Optional[bool] = None,
    tag_scope       : typ.Optional[str] = None,
    pre_commit_hook : typ.Optional[str] = None,
    post_commit_hook: typ.Optional[str] = None,
) -> config.Config:
    if commit is False and tag_commit:
        raise ValueError("--no-commit and --tag-commit cannot be used at the same time")
    if commit is False and push:
        raise ValueError("--no-commit and --push cannot be used at the same time")

    if commit is not None:
        cfg = cfg._replace(commit=commit)

    if not cfg.commit and tag_commit:
        raise ValueError("--tag-commit requires either --commit or commit=True in your config")
    if not cfg.commit and push:
        raise ValueError("--push requires either --commit or commit=True in your config")

    if tag_commit is not None:
        cfg = cfg._replace(tag=tag_commit)
    if push is not None:
        cfg = cfg._replace(push=push)
    if tag_scope is not None:
        cfg = cfg._replace(tag_scope=config.TagScope(tag_scope))
    if pre_commit_hook is not None:
        cfg = cfg._replace(pre_commit_hook=pre_commit_hook)
    if post_commit_hook is not None:
        cfg = cfg._replace(post_commit_hook=post_commit_hook)

    return cfg


def _sub_msg_template(message: str) -> str:
    return re.sub(r"\b(OLD|NEW)\b", r"{\1_VERSION}", message)


@cli.command()
@dry_option
@fetch_option
@verbose_option
@click.option(
    "--allow-dirty",
    default=False,
    is_flag=True,
    help=(
        "Commit even when working directory is has uncomitted changes. "
        "(WARNING: The commit will still be aborted if there are uncomitted "
        "to files with version strings."
    ),
)
@click.option(
    "--ignore-vcs-tag",
    default=False,
    is_flag=True,
    help="Ignore VCS tag invariant and update version anyway.",
)
@version_options
@click.option(
    "-c",
    "--commit-message",
    default=None,
    metavar="<TMPL>",
    help="Set commit message template.",
)
@click.option(
    "--tag-message",
    default=None,
    metavar="<TMPL>",
    help="Set tag message template.",
)
@click.option(
    "--commit/--no-commit",
    default=None,
    help="Create a commit with all updated files.",
)
@click.option(
    "--tag-commit/--no-tag-commit",
    default=None,
    help="Tag the newly created commit.",
)
@click.option(
    "--push/--no-push",
    default=None,
    help="Push to the default remote.",
)
@click.option(
    "--tag-scope",
    default=None,
    metavar="[default|global|branch]",
    type=click.Choice([e.value for e in config.TagScope]),
    help="Tag scope for the current version.",
)
@click.option(
    "--pre-commit-hook",
    type=click.Path(exists=True),
    metavar="<PATH>",
    help="Custom script that runs before the commit step",
)
@click.option(
    "--post-commit-hook",
    type=click.Path(exists=True),
    metavar="<PATH>",
    help="Custom script that runs after the commit step is completed",
)
def update(
    dry             : bool = False,
    allow_dirty     : bool = False,
    ignore_vcs_tag  : bool = False,
    fetch           : bool = True,
    verbose         : int = 0,
    major           : bool = False,
    minor           : bool = False,
    patch           : bool = False,
    tag             : typ.Optional[str] = None,
    tag_num         : bool = False,
    pin_increments  : bool = False,
    pin_date        : bool = False,
    date            : typ.Optional[str] = None,
    set_version     : typ.Optional[str] = None,
    commit_message  : typ.Optional[str] = None,
    tag_message     : typ.Optional[str] = None,
    commit          : typ.Optional[bool] = None,
    tag_commit      : typ.Optional[bool] = None,
    push            : typ.Optional[bool] = None,
    tag_scope       : typ.Optional[str] = None,
    pre_commit_hook : typ.Optional[str] = None,
    post_commit_hook: typ.Optional[str] = None,
) -> None:
    """Update project files with the incremented version string."""
    verbose = max(_VERBOSE, verbose)
    _configure_logging(verbose)
    _validate_release_tag(tag)
    maybe_date = _validate_date(date, pin_date)

    _, cfg = config.init(project_path=".")

    if cfg is None:
        logger.error("Could not parse configuration.")
        sys.exit(1)

    try:
        cfg = _parse_vcs_options(
            cfg, commit, tag_commit, push, tag_scope, pre_commit_hook, post_commit_hook
        )
    except ValueError as ex:
        logger.warning(f"Invalid argument: {ex}")
        sys.exit(1)

    if not ignore_vcs_tag:
        cfg = _update_cfg_from_vcs(cfg, fetch)

    old_version = cfg.current_version
    if set_version is None:
        new_version = incr_dispatch(
            old_version,
            raw_pattern=cfg.version_pattern,
            major=major,
            minor=minor,
            patch=patch,
            tag=tag,
            tag_num=tag_num,
            pin_increments=pin_increments,
            pin_date=pin_date,
            maybe_date=maybe_date,
        )
    else:
        new_version = set_version

    if new_version is None:
        _log_no_change('update', cfg.version_pattern)
        sys.exit(1)

    uniqueness_check = cfg.tag_scope == config.TagScope.BRANCH or set_version is not None

    if not _is_valid_version(
        cfg.version_pattern, old_version, new_version, unique=uniqueness_check
    ):
        if set_version:
            logger.error(f"Invalid argument --set-version='{set_version}'")
        sys.exit(1)

    logger.info(f"Old Version: {old_version}")
    logger.info(f"New Version: {new_version}")

    if dry or verbose >= 2:
        _print_diff(cfg, new_version)

    if commit_message is None:
        commit_msg_template = cfg.commit_message
    else:
        commit_msg_template = _sub_msg_template(commit_message)

    tag_msg_template = cfg.tag_message if tag_message is None else _sub_msg_template(tag_message)

    tag_and_commit_message_kwargs = {
        'new_version'       : new_version,
        'old_version'       : old_version,
        'NEW_VERSION'       : new_version,
        'OLD_VERSION'       : old_version,
        'new_version_pep440': version.to_pep440(new_version),
        'old_version_pep440': version.to_pep440(old_version),
    }

    try_commit_message = commit_msg_template.format(**tag_and_commit_message_kwargs)
    try_tag_message    = tag_msg_template.format(**tag_and_commit_message_kwargs)

    if dry:
        return

    _try_update(cfg, new_version, try_commit_message, try_tag_message, allow_dirty)


if __name__ == '__main__':
    cli()



# This file is part of the bumpver project
# https://github.com/mbarkhau/bumpver
#
# Copyright (c) 2018-2023 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import os
import sys
import logging
import subprocess as sp

from bumpver import pathlib as pl

logger = logging.getLogger("bumpver.hooks")


def run(path: str, old_version: str, new_version: str) -> None:
    env = dict(os.environ, BUMPVER_OLD_VERSION=old_version, BUMPVER_NEW_VERSION=new_version)

    try:
        # Python2 compatibility
        # pylint:disable=consider-using-with
        proc = sp.Popen(str(pl.Path(path).absolute()), env=env, stdout=sp.PIPE, stderr=sp.PIPE)
        if proc.stdout is not None:
            with proc.stdout as out:
                for line in iter(out.readline, b''):
                    logger.info(f"\t{line.decode('utf8').strip()}")
        if proc.stderr is not None:
            with proc.stderr as err:
                for line in iter(err.readline, b''):
                    logger.error(f"\t{line.decode('utf8').strip()}")
        proc.wait()
    except IOError as err:
        logger.error(f"\t{err}")
        logger.error("Script exited with an error. Stopping")
        sys.exit(1)

    if proc.returncode != 0:
        logger.error("Script exited with an error. Stopping")
        sys.exit(1)
